r"""
Old code that might moved or be removed.
"""
# ****************************************************************************
#       Copyright (C) 2011-2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

#
#from sage.structure.sage_object import SageObject
#
#@cached_function
#def _fact_and_irr(k):
#    r"""
#    intermediate function to accelerate the computation of
#    number_of_irreducible_permutations
#
#    the length of the result is k-1
#
#    [p(j)*factorial(k-j)]
#
#    TESTS::
#
#        sage: _fact_and_irr(1) == ()
#        True
#        sage: _fact_and_irr(2) == (1,)
#        True
#        sage: _fact_and_irr(3) == (2,1)
#        True
#    """
#    if k == 1: return ()
#    rec = _fact_and_irr(k-1)
#    return (number_of_irreducible_permutations(k-1),) + tuple((j+2)*rec[j] for j in range(k-2))
#
#@cached_function
#def number_of_irreducible_permutations(k):
#    r"""
#    Return the number of irreducible permutations on k letters.
#
#    Sloane A003319 (see also A158882, A167894)
#
#    EXAMPLES::
#
#        sage: for i in range(6): print i, number_of_irreducible_permutations(i)
#        0 1
#        1 1
#        2 1
#        3 3
#        4 13
#        5 71
#    """
#    if k <= 0: return 0
#    return (factorial(k) - sum(_fact_and_irr(k)))
#
#def rauzy_move((ltop,lbot),l):
#    r"""
#    Start from a generalized permutation given as ((ltop,lbot),l)
#    where ltop and lbot are lists with the top and bottom labels
#    (the labels are assumed to pair up -- no check), and l is a
#    dictionary of lengths for the labels used in ltop, lbot.
#
#    rauzy_move will apply one Rauzy move at the rightmost end:
#
#    - if rightmost top and bottom intervals have distinct lengths,
#      do the usual thing
#
#    - if rightmost top and bottom intervals have same label, error
#
#    - if rightmost top and bottom intervals have same length but
#      not same label, just remove one of the labels
#
#    Note: this code was originally used to compute the connected component of an
#    origami. This is no more ncessary but it might be a good idea to keep for
#    double check
#    """
#    if len(ltop) < 2 and len(lbot) < 2:
#        raise ValueError("cannot induce, too few intervals")
#    ktop = ltop[-1]
#    kbot = lbot[-1]
#    lltop = ltop[:-1]
#    llbot = lbot[:-1]
#    ll = l.copy()
#    if l[ktop] - l[kbot] > 0:
#        # print "bottom interval shorter\n"
#
#        # ltop[-1] is some symbol ktop, which appears twice: once as ltop[-1], and the other
#        # time either as ltop[j] for j < len(ltop) - 1 or as lbot[j] for j < len(lbot) - 1;
#        # as ltop[-1], ktop needs to be replaced by [ltop[-1],lbot[-1]],
#        # the other occurrence of ktop needs to be replaced
#        # - by [ltop[-1],lbot[-1]] if it is lbot[j]
#        # - by [lbot[-1],ltop[-1]] if it is ltop[j]
#        # the dictionary of lengths needs one update: ll[ktop] = l[ktop] - l[kbot]
#
#        if ktop in lltop: lltop.insert(lltop.index(ktop),kbot)
#        else: llbot.insert(llbot.index(ktop)+1,kbot)
#        lltop.append(ktop)
#        ll[ktop] = l[ktop] - l[kbot]
#
#    elif l[ktop] - l[kbot] < 0:
#        # print "top interval shorter\n"
#
#        # lbot[-1] is some symbol kbot, which appears twice: once as lbot[-1], and the other
#        # time either as lbot[j] for j < len(lbot) - 1 or as ltop[j] for j < len(ltop) - 1;
#        # as lbot[-1], kbot needs to be replaced by [lbot[-1],ltop[-1]],
#        # the other occurrence of ktop needs to be replaced
#        # - by [lbot[-1],ltop[-1]] if it is ltop[j]
#        # - by [ltop[-1],lbot[-1]] if it is lbot[j]
#        # the dictionary of lenghts needs one update: ll[kbot] = l[kbot] - l[ktop]
#
#        if kbot in llbot: llbot.insert(llbot.index(kbot),ktop)
#        else: lltop.insert(lltop.index(kbot)+1,ktop)
#        llbot.append(kbot)
#        ll[kbot] = l[kbot] - l[ktop]
#
#    elif ktop == kbot:
#        # in this case the surface has a torus component... weird.
#        raise ValueError("it seems the surface \n %s \n %s \n %s either is disconnected or has genus 1" %s(ltop,lbot,l))
#    else:
#        # ktop != kbot but the intervals labeled ktop and kbot have equal lengths
#        # just remove ktop and kbot at the end of ltop and lbot, and match the
#        # remaining ktop and kbot (keep only one of the names; also in the length dict)
#        if ktop in lltop: lltop[lltop.index(ktop)] = kbot
#        else: llbot[llbot.index(ktop)] = kbot
#        ll.pop(ktop)
#
#    return ((lltop,llbot),ll)
#
#def iet_of_origami(o):
#    r"""
#    x, y are permutations which together represent an origami
#    this origami is assumed to be a torus cover
#    (ie corresponds to an abelian differential)
#
#    First get a multi-iet by considering the first return to I, the union of antidiagonals
#    of all squares, of the flow in direction (1,u), where u is the golden mean.
#    The antidiagonals are parametrized by their horizontal coordinate, so that they appear
#    to have length 1, therefore I has length n. Apply Rauzy moves until I has length 1.
#
#    We could define an iet by using this permutation and this length vector
#    p = Permutation([(lambda k: 2*y((k+1)//2) if k%2 else 2*x(k//2)-1)(j) for j in range(1,n+1)])
#    l = [2-u,u-1] * n
#    """
#    from surface_dynamics.interval_exchanges.all import iet
#    from sage.rings.number_field.all import NumberField
#    from sage.rings.polynomial.all import PolynomialRing
#    from sage.rings.all import RR
#
#    sigma = o.r()
#    tau = o.u()
#    n = o.nb_squares()
#
#    R = PolynomialRing(QQ,'x')
#    x = R.gen()
#    K = NumberField(x**2 - x - 1, 'a',embedding=RR(1.618))
#    phi = K.gen()
#
#    ltop = []
#    for k in range(1,n+1):
#        ltop.extend([(k,0),(k,1)])
#    lbot = []
#    for k in range(1,n+1):
#        lbot.extend([(sigma.inverse()(k),1),(tau.inverse()(k),0)])
#    l = {}
#    for k in range(1,n+1):
#        l[k,0] = 2 - RR(phi)
#        l[k,1] = RR(phi) - 1
#
#    while sum(l.values())>1:
#        ((ltop, lbot), l) = rauzy_move((ltop,lbot),l)
#
#    return iet.Permutation(ltop,lbot),l
#
#    # chain complex of homology
#
#    @cached_method
#    def chain_complex(self,ring=None):
#        r"""
#        Return the chain complex associated to self
#        """
#        return SurfaceChainComplex(self,ring)
#
#

##############################
# Chain complex and homology #
##############################

#class SurfaceChainComplex(SageObject):
#    r"""
#    Surface chain complex associated to a Ribbon graph
#
#    -> intersection of cycles
#
#    EXAMPLES::
#
#        sage: g = RibbonGraph(edges='(0,1)(2,3)(4,5)(6,7)',faces='(0,2,4,6,1,3,5,7)')
#        sage: C = g.chain_complex()
#        sage: C.chain_space(0).rank() == g.num_vertices()
#        True
#        sage: C.chain_space(1).rank() == g.num_edges()
#        True
#        sage: C.chain_space(2).rank() == g.num_faces()
#        True
#
#        sage: C.cycle_space(0).rank() - C.boundary_space(0).rank() == 0
#        True
#        sage: C.cycle_space(1).rank() - C.boundary_space(1).rank() == 2 * g.genus()
#        True
#        sage: C.cycle_space(2).rank() - C.boundary_space(2).rank() == 1
#        True
#    """
#    def __init__(self, g, ring=None):
#        self._ribbon_graph = g
#        if ring is None:
#            ring = ZZ
#        self._base_ring = ring
#        self._chain_spaces = [
#            FreeModule(ring,g.num_vertices()),
#            FreeModule(ring,g.num_edges()),
#            FreeModule(ring,g.num_faces())]
#
#        self._derivatives = []
#        m = matrix(ring, [1]*g.num_vertices())
#        self._derivatives.append(m)
#
#        m = matrix(ring, g.num_vertices(), g.num_edges())
#        for i,e in enumerate(g.edges()):
#            m[g.dart_to_vertex(e[0]),i] -= 1
#            m[g.dart_to_vertex(e[1]),i] += 1
#        self._derivatives.append(m)
#
#        m = matrix(ring, g.num_edges(), g.num_faces())
#        for i,f in enumerate(g.faces()):
#            for e in f:
#                j,o = g.dart_to_edge(e,orientation=True)
#                m[j,i] += o
#        self._derivatives.append(m)
#
#    def parent(self):
#        r"""
#        Returns the ribbon graph from which this chain complex is defined.
#        """
#        return self._ribbon_graph
#
#    def _repr_(self):
#        return "Chain complex over %s" %str(self._base_ring)
#
#    def differential(self, degree=None):
#        r"""
#        Returns a differential of given degree.
#        """
#        if degree is None:
#            return self._derivatives
#        return self._derivatives[degree]
#
#    def chain_space(self, degree=None):
#        if degree is None:
#            return self._chain_spaces
#        return self._chain_spaces[degree]
#
#    @cached_method
#    def cycle_space(self,degree=None):
#        r"""
#        The space of cycles.
#        """
#        if degree is None:
#            return [self.cycle_space(i) for i in range(3)]
#        return self._derivatives[degree].right_kernel()
#
#    @cached_method
#    def boundary_space(self,degree=None):
#        r"""
#        Return the boundary space of the given ``degree``.
#        """
#        if degree is None:
#            return [self.boundary_space(i) for i in range(3)]
#        if degree == 2:
#            return self.chain_space(2).submodule([])
#        else:
#            return self._derivatives[degree+1].column_space()
#




r"""
Origami

An origami is:

- a couple of permutations

- a covering of the torus ramified over one point

- gluing of squares

- Abelian differential on Riemann surface with rational periods

EXAMPLES::

    sage: from surface_dynamics import *

    sage: E = origamis.EierlegendeWollmilchsau()
    sage: E.r()
    (1,2,3,4)(5,6,7,8)
    sage: E.u()
    (1,5,3,7)(2,8,4,6)
    sage: E.stratum_component()
    H_3(1^4)^c
    sage: E.lyapunov_exponents_approx()   # abs tol 1e-3
    [0.0000485630931783940, 0.0000452662733371477]

    sage: o = Origami('(1,2,3,4,5,6)','(1,7)')
    sage: V = o.veech_group()
    sage: V
    Arithmetic subgroup of index 54
    sage: G = V.coset_graph()
    sage: G.diameter()
    16
"""
# ****************************************************************************
#       Copyright (C) 2011-2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from __future__ import print_function, absolute_import
from six.moves import range, map, filter, zip
from six import iteritems

from surface_dynamics.flat_surfaces.origamis.origami_dense import Origami_dense_pyx
from surface_dynamics.misc.permutation import perm_init, perm_check

from sage.structure.sage_object import SageObject
from sage.groups.perm_gps.permgroup import PermutationGroup

try:
    # Trac #28652: Rework the constructor of PermutationGroupElement
    from sage.groups.perm_gps.constructor import PermutationGroupElement
except ImportError:
    from sage.groups.perm_gps.permgroup import PermutationGroupElement
from sage.misc.cachefunc import cached_method
from copy import copy
from sage.matrix.constructor import matrix, identity_matrix

from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets

from sage.rings.integer import Integer
from sage.modules.free_module import VectorSpace
from sage.rings.rational_field import QQ
from sage.structure.parent import Parent
from sage.structure.element import Element
from sage.structure.unique_representation import UniqueRepresentation

from sage.interfaces.gap import gap

from sage.plot.plot import options

def flatten_word(w):
    l = []
    for i,j in w:
        l.append(i*j)
    return ''.join(l)

def permutation_simplicial_action(r,u,n,w):
    r"""
    From a word in 'l','r' return the simplicial action on homology as
    well as the obtained origami.

    INPUT:

    - ``r`` and ``u`` - permutations

    - ``n`` - the degree of the permutations

    - ``w`` - a string in 'l' and 'r' or a list of 2-tuples which are made of
      the letter 'l' or 'r' and a positive integer

    """
    if w is None:
        w = []
    elif isinstance(w,list):
        w = flatten_word(w)

    res = identity_matrix(2*n)

    for letter in reversed(w):
        if letter == 'l':
            u = u*~r
            m = identity_matrix(2*n)
            m[:n,n:] = u.matrix()
            res = m * res
        elif letter == 'r':
            r = r*~u
            m = identity_matrix(2*n)
            m[n:,:n] = r.matrix()
            res = m * res
        else:
            raise ValueError("does not understand the letter %s" %str(letter))

    return r,u,res

#
# Origami and pillow case cover constructors
#

def Origami(r, u,
        sparse=False,
        check=True,
        as_tuple=False,
        positions=None, name=None):
    r"""
    Constructor for origami

    INPUT:

    - ``r``, ``u`` - two permutations

    - ``sparse`` - boolean (default: False)

    - ``check`` - boolean (default: True) - whether or not check the input

    - ``as_tuple`` - boolean (default: False) - if True, assume that ``r`` and
      ``u`` are tuples on [0,...,N-1] (time efficient)

    - ``positions`` - list of 2-tuples (default: None) - position of the squares
      for drawings

    - ``name`` - an optional name to the origami

    TESTS::

        sage: from surface_dynamics import Origami
        sage: Origami('(0)(1,3)(2)', '(0,3,2,1)', as_tuple=True)
        (1)(2,4)(3)
        (1,4,3,2)
        sage: Origami('(0)(1,3)(2)', '(0,3,2,1)', as_tuple=True)
        (1)(2,4)(3)
        (1,4,3,2)
        sage: Origami((0,3,2,1), [3,0,1,2], as_tuple=True)
        (1)(2,4)(3)
        (1,4,3,2)
        sage: Origami([0,3,2,1], "(0,3,2,1)", as_tuple=True)
        (1)(2,4)(3)
        (1,4,3,2)
    """
    if not as_tuple:
        r = PermutationGroupElement(r, check=check)
        u = PermutationGroupElement(u, check=check)

        r = [i-1 for i in r.domain()]
        u = [i-1 for i in u.domain()]

        N = max(len(r),len(u))
        r.extend(range(len(r),N))
        u.extend(range(len(u),N))

    elif check:
        if not isinstance(r, (tuple, list)):
            r = perm_init(r)
        if not isinstance(u, (tuple, list)):
            u = perm_init(u)
        N = len(r)
        if len(u) != N:
            raise ValueError("the two tuples must be of the same length")
        if not perm_check(r, N):
            raise ValueError("invalid permutation %s" % (r,))
        if not perm_check(u, N):
            raise ValueError("invalid permutation %s" % (u,))

    o = Origami_dense_pyx(tuple(r),tuple(u))

    if check and not o.is_connected():
        print("Warning: the origami is not connected")

    if name is not None:
        o.rename(name)
    if positions is not None:
        o.set_positions(positions)
    return o

#
# Other stuff which should be moved in a more geometrical place (like CW
# complex or embeeded graph or whatever).
#

from sage.categories.action import Action

class ActionOnOrigamiObjects(Action):
    r"""
    Generic action of the automorphism group of an origami.
    """
    def __init__(self, objects):
        import operator
        o = objects.origami()
        # Action.__init__(G,S,is_left,op)
        Action.__init__(
                self,
                o.automorphism_group(),
                objects,
                True,
                operator.mul)

    def _call_(self, g, x):
        r"""
        Action of g on x.
        """
        return x._acted_upon_(g, True)

class OrigamiObjects(Parent):
    def chain_space(self):
        return VectorSpace(QQ, self.cardinality())

    @cached_method
    def cycle_space(self):
        r"""
        The space of cycles.
        """
        return self.derivative().right_kernel()

    @cached_method
    def border_space(self):
        r"""
        The border space.
        """
        return self.derivative().column_space()

class OrigamiFaces(OrigamiObjects):
    def __init__(self, origami):
        Parent.__init__(self, category=FiniteEnumeratedSets())
        self._origami = origami

    def cardinality(self):
        return self._origami.nb_squares()

    def derivative(self):
        n = self._origami.nb_squares()
        ri = self._origami.r_inv_tuple()
        ui = self._origami.u_inv_tuple()
        der = matrix(2*n,n,ring=QQ)
        for i in range(n):
            if ui[i] != i: # horiz sides
                der[i,i] = -1
                der[ui[i],i] = 1
            if ri[i] != i: # vert sides
                der[n+i,i] = 1
                der[n+ri[i],i] = -1
        return der

class OrigamiEdges(OrigamiObjects):
    r"""
    """
    def __init__(self, origami):
        Parent.__init__(self)
        self._origami = origami
        ri = origami.r_inv_tuple()
        ui = origami.u_inv_tuple()
        n = origami.nb_squares()

        self._starts = [None]*(2*n)
        self._ends = [None]*(2*n)
        for i in range(n):
            self._ends[i] = origami.vertex(i+1).index()
            self._starts[i] = origami.vertex(ri[i]+1).index()

            self._ends[n+i] = origami.vertex(i+1).index()
            self._starts[n+i] = origami.vertex(ui[i]+1).index()

    def origami(self):
        return self._origami

    # about elements

    def _element_constructor(self, i):
        if i >= 0 and i < 2*o.nb_squares():
            return o.chain_complex().chain_space(1).gen(i)

    def start(self,i):
        return self._starts[i]

    def end(self,i):
        return self._ends[i]

    # about global derivation

    def cardinality(self):
        return 2*self._origami.nb_squares()

    def derivative(self):
        m = self._origami.nb_vertices(True)
        n = 2*self._origami.nb_squares()
        der = matrix(m,n,ring=QQ)
        for i in range(n):
            if self._starts[i] != self._ends[i]:
                der[self._ends[i], i] = 1
                der[self._starts[i], i] = -1
        return der

    # geometric stuff (should be moved in cylinder diagrams)

    def is_simple_closed_curve(self, c):
        r"""
        Test if c is a simple closed curve
        """
        assert(c in self.chain_space())
        d = c.dict()
        for i in d:
            if d[i] != 1 and d[i] != -1:
                return False

        vertices = set([])
        for i in d:
            if d[i] == 1:
                v = self.end(i)
            else:
                v = self.start(i)
            if v in vertices:
                return False

            vertices.add(v)

        return True

    def simple_closed_curve_to_vertex_in_out(self, c):
        r"""
        From a curve c (as a vector) returns two dictionnaries associated to
        input/output view from each visited vertices
        """
        c_in = {}
        c_out = {}
        for i,j in iteritems(c.dict()):
            if j == 1:
                c_in[self.start(i)] = i
                c_out[self.end(i)] = i
            elif j == -1:
                c_in[self.end(i)] = i
                c_out[self.start(i)] = i
            else:
                raise ValueError("not a simple closed curve")
        return c_in, c_out

    def basis_of_simple_closed_curves(self):
        from sage.graphs.digraph import DiGraph
        from sage.all import randint

        n = self._origami.nb_squares()
        C = self.chain_space()
        G = DiGraph(multiedges=True,loops=True,implementation='c_graph')

        for i in range(2*n):
            G.add_edge(self._starts[i], self._ends[i], i)

        waiting = [0]
        gens = []
        reps = [None] * G.num_verts()
        reps[0] = C.zero()

        while waiting:
            x = waiting.pop(randint(0,len(waiting)-1))
            for v0,v1,e in G.outgoing_edges(x):
                if reps[v1] is not None:
                    gens.append(reps[v0] + C.gen(e) - reps[v1])
                else:
                    reps[v1] = reps[v0] + C.gen(e)
                    waiting.append(v1)

        return gens

    def intersection(self, c1, c2):
        r"""
        Returns the intersection of c1 and c2 assuming that they are simple
        closed curves
        """
        assert(self.is_simple_closed_curve(c1) and self.is_simple_closed_curve(c2))

        c1_in, c1_out = self.simple_closed_curve_to_vertex_in_out(c1)
        c2_in, c2_out = self.simple_closed_curve_to_vertex_in_out(c2)

        intersection = 0
        for vert in c1_in:
            if vert in c2_in:
                if c1_in[vert] == c2_in[vert]:
                    if c1_out[vert] == c2_out[vert]:
                        pass
                    elif are_cyclically_ordered(
                            c1_in[vert], c1_out[vert], c2_out[vert]):
                        intersection += 1

                elif c1_out[vert] == c2_out[vert]:
                    if are_cyclically_ordered(
                            c1_out[vert], c2_in[vert], c1_in[vert]):
                        intersection -= 1

                elif are_cyclically_ordered(
                        c1_in[vert],c2_in[vert],c1_out[vert]):
                    if are_cyclically_ordered(
                            c1_out[vert],c2_out[vert],c1_in[vert]):
                            intersection += 1

                elif are_cyclically_ordered(
                        c1_in[vert],c2_out[vert],c1_out[vert]):
                    intersection -= 1

        return intersection

    def winding(self, c):
        r"""
        Return the winding of the curve c
        """
        assert(self.is_simple_closed_curve(c))
        pass

class OrigamiVertices(OrigamiObjects):
    r"""
    The set of vertices of an origami.

    It is in bijection with the biclasses H  G  C where H is the stabilizer of
    1 and C is the subgroup generated by the commutator ``r u r**-1 u **-1``.
    """
    def __init__(self, origami, register_automorphism_action=False):
        Parent.__init__(self, category=FiniteEnumeratedSets())
        self._origami = origami

        self._vertices = []
        self._vertices_inv = {}
        self._vertex_from_ur = {}
        self._vertex_from_dl = {}

        r = origami.r()
        u = origami.u()
        ru = r*u
        vperm = ru*~r*~u

        for ur in vperm.cycle_tuples(singletons=True):
            dl = tuple(map(ru,ur))
            v = OrigamiVertex(self, ur, dl)
            self._vertices_inv[v] = len(self._vertices)
            self._vertices.append(v)
            for i in ur:
                self._vertex_from_ur[i] = v
            for i in dl:
                self._vertex_from_dl[i] = v

        if register_automorphism_action:
            self.register_action(ActionOnOrigamiObjects(self))

    def origami(self):
        r"""
        Return the underlying origami.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: o = Origami('(1,2,3,4)','(1,5)')
            sage: V = o._vertices()
            sage: V.origami()
            (1,2,3,4)(5)
            (1,5)(2)(3)(4)
            sage: V.origami() is o
            True
        """
        return self._origami

    def cardinality(self):
        r"""
        The number of vertices.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: o = Origami('(1,2,3,4)','(1,5)')
            sage: V = o._vertices()
            sage: V.cardinality()
            3
        """
        return Integer(len(self._vertices))

    def chain_space(self):
        r"""
        Return the space of chain on this finite set of vertices.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: o = Origami('(1,2,3,4)', '(1,5)')
            sage: V = o._vertices()
            sage: V.chain_space()
            Vector space of dimension 3 over Rational Field
        """
        return VectorSpace(QQ, self.cardinality())

    def derivative(self):
        r"""
        Return the derivative matrix.

        That is a matrix from the chain space to `\QQ`.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: o = Origami('(1,2,3,4)', '(1,5)')
            sage: V = o._vertices()
            sage: V.derivative()
            [1 1 1]
        """
        return matrix([1]*len(self._vertices), ring=QQ)

    def _repr_(self):
        return "Vertices of origami\n%s" %self.origami()

    def __contains__(self, v):
        return v in self._vertices

    def up_right_vertex(self, i):
        r"""
        Return the vertex that is in the up right corner of the ``i``-th square.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: o = Origami('(1,2,3,4)', '(1,5)')
            sage: V = o._vertices()
            sage: V.up_right_vertex(2)
            vertex (2,)
        """
        return self._vertex_from_ur[i]

    def down_left_vertex(self, i):
        r"""
        Return the vertex that is in the down left corner of the ``i``-th square.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: o = Origami('(1,2,3,4)', '(1,5)')
            sage: V = o._vertices()
            sage: V.down_left_vertex(2)
            vertex (1, 5, 4)
        """
        return self._vertex_from_dl[i]

    vertex = up_right_vertex
    _element_constructor_ = up_right_vertex

    def vertex_index(self, v):
        r"""
        Return the index of the vertex ``v``.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: o = Origami('(1,2,3,4)', '(1,5)')
            sage: V = o._vertices()
            sage: v0 = V.up_right_vertex(2)
            sage: v0
            vertex (2,)
            sage: V.vertex_index(v0)
            1
            sage: V[1] == v0
            True

            sage: v1 = V.up_right_vertex(5)
            sage: v1
            vertex (1, 5, 4)
            sage: V.vertex_index(v1)
            0
            sage: V[0] == v1
            True
        """
        return self._vertices_inv[v]

    def __getitem__(self, i):
        r"""
        Access to a given element.
        """
        return self._vertices[i]

    def __iter__(self):
        r"""
        Iterator over the vertices.
        """
        return iter(self._vertices)

class OrigamiVertex(Element):
    def __init__(self, parent, ur, dl):
        Element.__init__(self, parent)
        self._ur = ur
        self._dl = dl

    def _repr_(self):
        return "vertex %s" %(str(self._ur))

    def __str__(self):
        return str(self._ur)

    def __hash__(self):
        return hash((self._ur,self._dl))

    def __eq__(self, other):
        return (type(self) == type(other) and
                self.parent() == other.parent() and
                self._ur == other._ur)

    def __ne__(self, other):
        return (type(self) != type(other) or
                self.parent() != other.parent() or
                self._ur != other._ur)

    def _acted_upon_(self, g, on_left):
        r"""
        Return the action of the element g of the automorphism group of the
        origami on self.
        """
        if not on_left:
            return self.parent().up_right_vertex(g(self._ur[0]))
        else:
            return self.parent().up_right_vertex(g(self._ur[0]))

    def index(self):
        return self.parent().vertex_index(self)

    def degree(self):
        return len(self._ur)-1

    def up_right_tuple(self):
        return self._ur

    def down_left_tuple(self):
        return self._dl

    # adjacent vectors

    def adjacent_edge_indices(self):
        r"""
        Returns a 2-tuple (incoming,outgoing)
        """
        return self.incoming_edge_indices(), self.outgoing_edge_indices()

    def outgoing_edge_indices(self):
        n = self.parent().origami().nb_squares()
        ri = ~self.parent().origami().r()
        ui = ~self.parent().origami().u()
        res = []
        for i in self._dl:
            res.append(ui(i)-1)
            res.append(n+ri(i)-1)
        return res

    def incoming_edge_indices(self):
        res = []
        n = self.parent().origami().nb_squares()
        for i in self._ur:
            res.append(i-1)
            res.append(n+i-1)
        return res

    @cached_method
    def edge_positions(self):
        r"""
        The position of the edges.
        """
        d_in = {}
        d_out = {}
        e_in = self.incoming_edge_indices()
        e_out = self.outgoing_edge_indices()
        k = 0
        for i in range(0,len(e_in),2):
            d_in[e_in[i]] = k
            d_in[e_in[i+1]] = k+1
            d_out[e_out[i]] = k+2
            d_out[e_out[i+1]] = k+3
            k += 4
        return d_in, d_out

    def incoming_edge_position(self, i):
        return self.edge_positions()[0][i]

    def outgoing_edge_positions(self, i):
        return self.edge_positions()[1][i]

class OrigamiChainComplex(SageObject, UniqueRepresentation):
    r"""
    Chain complex for reduced homology of the origami
    """
    def __init__(self, origami):
        self._origami = origami
        self._objects = [OrigamiVertices(origami), OrigamiEdges(origami), OrigamiFaces(origami)]

    def origami(self):
        r"""
        Return the underlying origami.
        """
        return self._origami

    def chain_space(self, degree):
        r"""
        Chain space.

        INPUT:

        - ``degree`` -- either

        """
        assert(degree > -2 and degree < 3)
        if degree == -1:
            return VectorSpace(QQ,1)
        return self._objects[degree].chain_space()

    def cycle_space(self, degree):
        r"""
        Returns the space of cycles of degree i

        Zi = ker(der: C_i -> C_{i-1})
        """
        assert(degree > -2 and degree < 3)
        if degree == -1:
            return VectorSpace(QQ,1)
        return self._objects[degree].cycle_space()

    def border_space(self, degree):
        r"""
        Returns the space of borders of degree i

        Bi = im(der: C_{i+1} -> Ci)
        """
        assert(degree > -2 and degree < 3)
        if degree == 2:
            return self._objects[2].chain_space().subspace([])
        return self._objects[degree+1].border_space()

    def _repr_(self):
        return "Chain complex of origami\n%s" %(self.origami())



r"""
Labelled permutations

A labelled (generalized) permutation is better suited to study the dynamic of
a translation surface than a reduced one (see the module
:mod:`surface_dynamics.interval_exchanges.reduced`). The latter is more adapted to the
study of strata. This kind of permutation was introduced by Yoccoz [Yoc05]_
(see also [MMY03]_).

In fact, there is a geometric counterpart of labelled permutations. They
correspond to translation surface with marked outgoing separatrices (i.e. we fi
a label for each of them).

Remarks that Rauzy diagram of reduced objects are significantly smaller than
the one for labelled object (for the permutation a b d b e / e d c a c the
labelled Rauzy diagram contains 8760 permutations, and the reduced only 73).
But, as it is in geometrical way, the labelled Rauzy diagram is a covering of
the reduced Rauzy diagram.

AUTHORS:

- Vincent Delecroix (2009-09-29) : initial version

- Vincent Delecroix (2010-02-11) : correction and simplification of datatypes

TESTS::

    sage: from surface_dynamics.interval_exchanges.labelled import LabelledPermutationIET
    sage: LabelledPermutationIET([['a','b','c'],['c','b','a']])
    a b c
    c b a
    sage: LabelledPermutationIET([[1,2,3,4],[4,1,2,3]])
    1 2 3 4
    4 1 2 3
    sage: from surface_dynamics.interval_exchanges.labelled import LabelledPermutationLI
    sage: LabelledPermutationLI([[1,1],[2,2,3,3,4,4]])
    1 1
    2 2 3 3 4 4
    sage: LabelledPermutationLI([['a','a','b','b','c','c'],['d','d']])
    a a b b c c
    d d
    sage: from surface_dynamics.interval_exchanges.labelled import FlippedLabelledPermutationIET
    sage: FlippedLabelledPermutationIET([[1,2,3],[3,2,1]],flips=[1,2])
    -1 -2  3
     3 -2 -1
    sage: FlippedLabelledPermutationIET([['a','b','c'],['b','c','a']],flips='b')
     a -b  c
    -b  c  a
    sage: from surface_dynamics.interval_exchanges.labelled import FlippedLabelledPermutationLI
    sage: FlippedLabelledPermutationLI([[1,1],[2,2,3,3,4,4]], flips=[1,4])
    -1 -1
     2  2  3  3 -4 -4
    sage: FlippedLabelledPermutationLI([['a','a','b','b'],['c','c']],flips='ac')
    -a -a  b  b
    -c -c
    sage: from surface_dynamics.interval_exchanges.labelled import LabelledRauzyDiagram
    sage: p = LabelledPermutationIET([[1,2,3],[3,2,1]])
    sage: d1 = LabelledRauzyDiagram(p)
    sage: p = LabelledPermutationIET([['a','b'],['b','a']])
    sage: d = p.rauzy_diagram()
    sage: g1 = d.path(p, 'top', 'bottom')
    sage: g1.matrix()
    [1 1]
    [1 2]
    sage: g2 = d.path(p, 'bottom', 'top')
    sage: g2.matrix()
    [2 1]
    [1 1]
    sage: p = LabelledPermutationIET([['a','b','c','d'],['d','c','b','a']])
    sage: d = p.rauzy_diagram()
    sage: g = d.path(p, 't', 't', 'b', 't', 'b', 'b', 't', 'b')
    sage: g
    Path of length 8 in a Rauzy diagram
    sage: g.is_loop()
    True
    sage: g.is_full()
    True
    sage: s1 = g.orbit_substitution()
    sage: print(s1)
    a->adbd, b->adbdbd, c->adccd, d->adcd
    sage: s2 = g.interval_substitution()
    sage: print(s2)
    a->abcd, b->bab, c->cdc, d->dcbababcd
    sage: s1.incidence_matrix() == s2.incidence_matrix().transpose()
    True

REFERENCES:

.. [Yoc05] Jean-Cristophe Yoccoz "Echange d'Intervalles", Cours au college de
   France

.. [MMY03] Jean-Cristophe Yoccoz, Stefano Marmi and Pierre Moussa "On the
   cohomological equation for interval exchange maps", arXiv:math/0304469v1
"""
#*****************************************************************************
#       Copyright (C) 2008 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from __future__ import print_function, absolute_import
from six.moves import range, map, filter, zip

from sage.structure.sage_object import SageObject
from sage.misc.lazy_attribute import lazy_attribute

from copy import copy

import time
import surface_dynamics.interval_exchanges.lyapunov_exponents as lyapunov_exponents  # the cython bindings

from sage.combinat.words.alphabet import Alphabet, OrderedAlphabet
from sage.combinat.words.morphism import WordMorphism

from sage.matrix.constructor import Matrix, identity_matrix
from sage.rings.integer import Integer
from sage.combinat.words.alphabet import Alphabet
from sage.rings.infinity import Infinity

from .template import (OrientablePermutationIET, OrientablePermutationLI,
                       FlippedPermutationIET, FlippedPermutationLI,
                       RauzyDiagram, FlippedRauzyDiagram,
                       interval_conversion, side_conversion)

class LabelledPermutation(SageObject):
    r"""
    General template for labelled objects.

    .. WARNING::

       Internal class! Do not use directly!
    """
    def __getitem__(self, i):
        r"""
        TESTS::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation([0,1,2,3],[3,2,1,0])
            sage: p[0][0]
            0
            sage: p[1][2]
            1
            sage: p = iet.Permutation('a b c','c b a')
            sage: p[0]
            ['a', 'b', 'c']
            sage: p[1]
            ['c', 'b', 'a']

            sage: p = iet.Permutation('a b', 'b a', flips='a')
            sage: p[0]
            ['a', 'b']
            sage: p = iet.GeneralizedPermutation('c p p', 't t c', flips='ct')
            sage: p[1]
            ['t', 't', 'c']
        """
        return list(map(self._alphabet.unrank, self._labels[i]))

    def list(self, flips=False):
        r"""
        Returns a list of two lists corresponding to the intervals.

        INPUT:

        - ``flips`` - boolean (default: False) - if ``True`` returns instead of
          letters use pair of letter and flip.

        OUTPUT: two lists of labels (or labels with flips)

        EXAMPLES::

            sage: from surface_dynamics import *

        The list of an permutation from iet::

            sage: p1 = iet.Permutation('1 2 3', '3 1 2')
            sage: p1.list()
            [['1', '2', '3'], ['3', '1', '2']]
            sage: p1.alphabet("abc")
            sage: p1.list()
            [['a', 'b', 'c'], ['c', 'a', 'b']]

        Recovering the permutation from this list (and the alphabet)::

            sage: q1 = iet.Permutation(p1.list(),alphabet=p1.alphabet())
            sage: p1 == q1
            True

        The list of a quadratic permutation::

            sage: p2 = iet.GeneralizedPermutation('g o o', 'd d g')
            sage: p2.list()
            [['g', 'o', 'o'], ['d', 'd', 'g']]

        Recovering the permutation::

            sage: q2 = iet.GeneralizedPermutation(p2.list(),alphabet=p2.alphabet())
            sage: p2 == q2
            True

        Some non-orientable examples::

            sage: p = iet.GeneralizedPermutation('0 0 1 2 2 1', '3 3', flips='1')
            sage: p.list(flips=True)
            [[('0', 1), ('0', 1), ('1', -1), ('2', 1), ('2', 1), ('1', -1)], [('3', 1), ('3', 1)]]
            sage: p.list(flips=False)
            [['0', '0', '1', '2', '2', '1'], ['3', '3']]

            sage: iet.Permutation('a b c', 'c b a').list(flips=True)
            [[('a', 1), ('b', 1), ('c', 1)], [('c', 1), ('b', 1), ('a', 1)]]

        The list can be used to reconstruct the permutation::

            sage: p = iet.Permutation('a b c','c b a',flips='ab')
            sage: p == iet.Permutation(p.list(), flips=p.flips())
            True

        ::

            sage: p = iet.GeneralizedPermutation('a b b c','c d d a',flips='ad')
            sage: p == iet.GeneralizedPermutation(p.list(), flips=p.flips())
            True
        """
        if flips:
            if self._flips is None:
                flips = [[1] * len(self._labels[0]), [1] * len(self._labels[1])]
            else:
                flips = self._flips
            a0 = list(zip(map(self._alphabet.unrank, self._labels[0]), flips[0]))
            a1 = list(zip(map(self._alphabet.unrank, self._labels[1]), flips[1]))
        else:
            a0 = list(map(self._alphabet.unrank, self._labels[0]))
            a1 = list(map(self._alphabet.unrank, self._labels[1]))

        return [a0,a1]

    def rauzy_move_matrix(self, winner=None, side='right'):
        r"""
        Returns the Rauzy move matrix.

        This matrix corresponds to the action of a Rauzy move on the vector of
        lengths. By convention (to get a positive matrix), the matrix is define
        as the inverse transformation on the length vector.

        OUTPUT:

        matrix -- a square matrix of positive integers

        EXAMPLES:

            sage: from surface_dynamics import *

        ::

            sage: p = iet.Permutation('a b','b a')
            sage: p.rauzy_move_matrix('t')
            [1 0]
            [1 1]
            sage: p.rauzy_move_matrix('b')
            [1 1]
            [0 1]

        ::

            sage: p = iet.Permutation('a b c d','b d a c')
            sage: q = p.left_right_inverse()
            sage: m0 = p.rauzy_move_matrix(winner='top',side='right')
            sage: n0 = q.rauzy_move_matrix(winner='top',side='left')
            sage: m0 == n0
            True
            sage: m1 = p.rauzy_move_matrix(winner='bottom',side='right')
            sage: n1 = q.rauzy_move_matrix(winner='bottom',side='left')
            sage: m1 == n1
            True
        """
        if winner is None and side is None:
            return identity_matrix(len(self))

        winner = interval_conversion(winner)
        side = side_conversion(side)

        winner_letter = self._labels[winner][side]
        loser_letter = self._labels[1-winner][side]

        m = copy(identity_matrix(len(self)))
        m[winner_letter, loser_letter] = 1

        return m

    def rauzy_move_winner(self,winner=None,side=None):
        r"""
        Returns the winner of a Rauzy move.

        INPUT:

        - ``winner`` - either 'top' or 'bottom' ('t' or 'b' for short)

        - ``side`` - either 'left' or 'right' ('l' or 'r' for short)

        OUTPUT:

        -- a label

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b c d','b d a c')
            sage: p.rauzy_move_winner('top','right')
            'd'
            sage: p.rauzy_move_winner('bottom','right')
            'c'
            sage: p.rauzy_move_winner('top','left')
            'a'
            sage: p.rauzy_move_winner('bottom','left')
            'b'

        ::

            sage: p = iet.GeneralizedPermutation('a b b c','d c a e d e')
            sage: p.rauzy_move_winner('top','right')
            'c'
            sage: p.rauzy_move_winner('bottom','right')
            'e'
            sage: p.rauzy_move_winner('top','left')
            'a'
            sage: p.rauzy_move_winner('bottom','left')
            'd'
        """
        if winner is None and side is None:
            return None

        winner = interval_conversion(winner)
        side = side_conversion(side)

        return self[winner][side]

    def rauzy_move_loser(self,winner=None,side=None):
        r"""
        Returns the loser of a Rauzy move

        INPUT:

        - ``winner`` - either 'top' or 'bottom' ('t' or 'b' for short)

        - ``side`` - either 'left' or 'right' ('l' or 'r' for short)

        OUTPUT:

        -- a label

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b c d','b d a c')
            sage: p.rauzy_move_loser('top','right')
            'c'
            sage: p.rauzy_move_loser('bottom','right')
            'd'
            sage: p.rauzy_move_loser('top','left')
            'b'
            sage: p.rauzy_move_loser('bottom','left')
            'a'
        """
        if winner is None and side is None:
            return None

        winner = interval_conversion(winner)
        side = side_conversion(side)

        return self[1-winner][side]

def LabelledPermutationsIET_iterator(
    nintervals=None,
    irreducible=True,
    alphabet=None):
    r"""
    Returns an iterator over labelled permutations.

    INPUT:

    - ``nintervals`` - integer or None

    - ``irreducible`` - boolean (default: True)

    - ``alphabet`` - something that should be converted to an alphabet of at least nintervals letters

    OUTPUT:

    iterator -- an iterator over permutations

    TESTS::

        sage: from surface_dynamics import iet

        sage: for p in sorted(iet.Permutations_iterator(2, alphabet="ab")):
        ....:     print("%s\n****" % p)  #indirect doctest
        a b
        b a
        ****
        b a
        a b
        ****
        sage: for p in iet.Permutations_iterator(3, alphabet="abc"):
        ....:     print("%s\n*****" %p)   #indirect doctest
        a b c
        b c a
        *****
        a b c
        c a b
        *****
        a b c
        c b a
        *****
        a c b
        b a c
        *****
        a c b
        b c a
        *****
        a c b
        c b a
        *****
        b a c
        a c b
        *****
        b a c
        c a b
        *****
        b a c
        c b a
        *****
        b c a
        a b c
        *****
        b c a
        a c b
        *****
        b c a
        c a b
        *****
        c a b
        a b c
        *****
        c a b
        b a c
        *****
        c a b
        b c a
        *****
        c b a
        a b c
        *****
        c b a
        a c b
        *****
        c b a
        b a c
        *****
    """
    from itertools import product
    from sage.combinat.permutation import Permutations

    if irreducible is False:
        if nintervals is None:
            raise ValueError("choose a number of intervals")
        else:
            assert(isinstance(nintervals,(int,Integer)))
            assert(nintervals > 0)

            f = lambda x: LabelledPermutationIET([list(x[0]),list(x[1])],alphabet=alphabet,reduced=False)

            alphabet = Alphabet(alphabet)
            g = lambda x: [alphabet.unrank(k-1) for k in x]
            P = map(g, Permutations(nintervals))
            return map(f,product(P,repeat=2))
    else:
        return filter(
            lambda x: x.is_irreducible(),
            LabelledPermutationsIET_iterator(nintervals,False,alphabet))

class LabelledPermutationIET(LabelledPermutation, OrientablePermutationIET):
    """
    Labelled permutation for iet

    EXAMPLES::

        sage: from surface_dynamics import *

    Reducibility testing::

        sage: p = iet.Permutation('a b c', 'c b a')
        sage: p.is_irreducible()
        True

        sage: q = iet.Permutation('a b c d', 'b a d c')
        sage: q.is_irreducible()
        False

    Rauzy movability and Rauzy move::

        sage: p = iet.Permutation('a b c', 'c b a')
        sage: p.has_rauzy_move('top')
        True
        sage: p.rauzy_move('bottom')
        a c b
        c b a
        sage: p.has_rauzy_move('top')
        True
        sage: p.rauzy_move('top')
        a b c
        c a b

    Rauzy diagram::

        sage: p = iet.Permutation('a b c', 'c b a')
        sage: d = p.rauzy_diagram()
        sage: p in d
        True
    """
    def reduced(self):
        r"""
        Returns the associated reduced abelian permutation.

        OUTPUT:

        a reduced permutation -- the underlying reduced permutation


        EXAMPLES:

            sage: from surface_dynamics import *

            sage: p = iet.Permutation("a b c d","d c a b")
            sage: q = iet.Permutation("a b c d","d c a b",reduced=True)
            sage: p.reduced() == q
            True
        """
        from .reduced import ReducedPermutationIET

        return ReducedPermutationIET(self.list(), alphabet=self._alphabet, reduced=True)

    def rauzy_move_interval_substitution(self,winner=None,side=None):
        r"""
        Returns the interval substitution associated.

        INPUT:

        - ``winner`` - the winner interval ('top' or 'bottom')

        - ``side`` - (default: 'right') the side ('left' or 'right')

        OUTPUT:

        WordMorphism -- a substitution on the alphabet of the permutation

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b','b a')
            sage: print(p.rauzy_move_interval_substitution('top','right'))
            a->a, b->ba
            sage: print(p.rauzy_move_interval_substitution('bottom','right'))
            a->ab, b->b
            sage: print(p.rauzy_move_interval_substitution('top','left'))
            a->ba, b->b
            sage: print(p.rauzy_move_interval_substitution('bottom','left'))
            a->a, b->ab
        """
        d = dict([(letter,[letter]) for letter in self.letters()])

        if winner is None and side is None:
            return WordMorphism(d)

        winner = interval_conversion(winner)
        side = side_conversion(side)

        winner_letter = self.rauzy_move_winner(winner,side)
        loser_letter = self.rauzy_move_loser(winner,side)

        if side == 0:
            d[winner_letter] = [loser_letter,winner_letter]
        else:
            d[winner_letter] = [winner_letter,loser_letter]

        return WordMorphism(d)

    def rauzy_move_orbit_substitution(self,winner=None,side=None):
        r"""
        Return the action fo the rauzy_move on the orbit.

        INPUT:

        - ``i`` - integer

        - ``winner`` - the winner interval ('top' or 'bottom')

        - ``side`` - (default: 'right') the side ('right' or 'left')

        OUTPUT:

        WordMorphism -- a substitution on the alphabet of self

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b','b a')
            sage: print(p.rauzy_move_orbit_substitution('top','right'))
            a->ab, b->b
            sage: print(p.rauzy_move_orbit_substitution('bottom','right'))
            a->a, b->ab
            sage: print(p.rauzy_move_orbit_substitution('top','left'))
            a->a, b->ba
            sage: print(p.rauzy_move_orbit_substitution('bottom','left'))
            a->ba, b->b

        TESTS::

            sage: p = iet.Permutation('a1 a2', 'a2 a1')
            sage: p.rauzy_move_orbit_substitution('top','right').codomain().alphabet()
            {'a1', 'a2'}
        """
        d = dict([(letter,[letter]) for letter in self.letters()])

        if winner is None and side is None:
            return WordMorphism(d)

        winner = interval_conversion(winner)
        side = side_conversion(side)

        loser_letter = self.rauzy_move_loser(winner,side)

        top_letter = self.alphabet().unrank(self._labels[0][side])
        bottom_letter = self.alphabet().unrank(self._labels[1][side])

        d[loser_letter] = [bottom_letter,top_letter]

        return WordMorphism(d)

    def rauzy_diagram(self, **args):
        """
        Returns the associated Rauzy diagram.

        For more information try help(iet.RauzyDiagram).

        OUTPUT:

        Rauzy diagram -- the Rauzy diagram of the permutation

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b c', 'c b a')
            sage: d = p.rauzy_diagram()
        """
        return LabelledRauzyDiagram(self, **args)

    def suspension_cone(self, winner=None):
        r"""
        Return the cone of suspension data.

        A suspension data `\tau` for a permutation `(\pi_{top}, \pi_{bot})`
        on the alphabet `\mathcal{A}` is a real vector in `RR^\mathcal{A}`
        so that

        .. MATH::

            \forall 1 \leq k < d,\,
            \sum_{\beta: \pi_{top}(\beta) \leq k} \tau_\beta > 0
            \quad \text{and} \quad
            \sum_{\beta: \pi_{bot}(\beta) \leq k} \tau_\beta < 0.

        A suspension data determines half of a zippered rectangle construction.
        The other half is the length data that is a positive vector in
        `\RR^\mathcal{A}`.

        INPUT:

        - ``winner`` - (optional) either ``None``, ``"top"`` or ``"bottom"``. If
          not ``None`` , then return only half of the suspension cone corresponding
          to data that either comes from a top or bottom Rauzy induction.

        .. SEEALSO::

            :meth:`heights_cone`

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p = iet.Permutation('a b c d e f', 'e c b f d a')
            sage: H = p.suspension_cone()
            sage: H.dimension()
            6
            sage: rays = [r.vector() for r in H.rays()]
            sage: r = sum(randint(1,5)*ray for ray in rays)
            sage: r[0]>0 and r[0]+r[1] > 0 and r[0]+r[1]+r[2] > 0
            True
            sage: r[0]+r[1]+r[2]+r[3]>0
            True
            sage: r[0]+r[1]+r[2]+r[3]+r[4]>0
            True
            sage: r[4]<0 and r[4]+r[2]<0 and r[4]+r[2]+r[1] < 0
            True
            sage: r[4]+r[2]+r[1]+r[5]<0
            True
            sage: r[4]+r[2]+r[1]+r[5]+r[3]<0
            True
        """
        n = len(self)
        ieqs = []

        for i in range(1,len(self)):
            ieq = [0]*(n+1)
            for j in range(i):
                ieq[self._labels[0][j]+1] = 1
            ieqs.append(ieq)

            ieq = [0]*(n+1)
            for j in range(i):
                ieq[self._labels[1][j]+1] = -1
            ieqs.append(ieq)

        if winner is not None:
            winner = interval_conversion(winner)
            if winner == 0:
                ieqs.append([0] + [1] * len(self))
            elif winner == 1:
                ieqs.append([0] + [-1] * len(self))

        from sage.geometry.polyhedron.constructor import Polyhedron
        return Polyhedron(ieqs=ieqs)

    def heights_cone(self, side=None):
        r"""
        Return the cone of heights data.

        .. SEEALSO::

            :meth:`suspension_cone`

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p = iet.Permutation('a b c d', 'd c b a')
            sage: C = p.heights_cone()
            sage: C
            A 4-dimensional polyhedron in QQ^4 defined as the convex hull of 1 vertex and 5 rays
            sage: C.rays_list()
            [[0, 0, 1, 1], [0, 1, 1, 0], [0, 1, 1, 1], [1, 1, 0, 0], [1, 1, 1, 0]]

            sage: p.heights_cone('top').rays_list()
            [[0, 0, 1, 1], [0, 1, 1, 0], [0, 1, 1, 1], [1, 1, 0, 0]]
            sage: p.heights_cone('bot').rays_list()
            [[0, 0, 1, 1], [0, 1, 1, 0], [1, 1, 0, 0], [1, 1, 1, 0]]
        """
        I = self.intersection_matrix()
        C = self.suspension_cone(side)

        from sage.geometry.polyhedron.constructor import Polyhedron
        return Polyhedron(rays=[-I*c.vector() for c in C.rays()])

    def lyapunov_exponents_approx(self, nb_vectors=None, nb_experiments=10,
                                  nb_iterations=65536, return_speed=False,
                                  verbose=False, output_file=None):
        r"""
        Return approximate Lyapunov exponents of the KZ-cocycle.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation([1,2,3],[3,2,1])
            sage: p.lyapunov_exponents_approx()  # abs tol .05
            [1.000]
        """
        if self._flips:
            raise NotImplementedError("Lyapunov exponents not implemented for permutations with flips")
        c = self.cover([[0]]*len(self), as_tuple=True)
        return c.lyapunov_exponents_H_plus(
                    nb_vectors=nb_vectors, nb_experiments=nb_experiments,
                    nb_iterations=nb_iterations, return_speed=return_speed,
                    verbose=verbose, output_file=output_file)

class LabelledPermutationLI(LabelledPermutation, OrientablePermutationLI):
    r"""
    Labelled quadratic (or generalized) permutation

    EXAMPLES::

        sage: from surface_dynamics import *

    Reducibility testing::

        sage: p = iet.GeneralizedPermutation('a b b', 'c c a')
        sage: p.is_irreducible()
        True

    Reducibility testing with associated decomposition::

        sage: p = iet.GeneralizedPermutation('a b c a', 'b d d c')
        sage: p.is_irreducible()
        False
        sage: test, decomposition = p.is_irreducible(return_decomposition = True)
        sage: test
        False
        sage: decomposition
        (['a'], ['c', 'a'], [], ['c'])

    Rauzy movability and Rauzy move::

        sage: p = iet.GeneralizedPermutation('a a b b c c', 'd d')
        sage: p.has_rauzy_move(0)
        False
        sage: p.has_rauzy_move(1)
        True
        sage: q = p.rauzy_move(1)
        sage: q
        a a b b c
        c d d
        sage: q.has_rauzy_move(0)
        True
        sage: q.has_rauzy_move(1)
        True

    Rauzy diagrams::

        sage: p = iet.GeneralizedPermutation('0 0 1 1','2 2')
        sage: r = p.rauzy_diagram()
        sage: p in r
        True
    """
    def has_right_rauzy_move(self, winner):
        r"""
        Test of Rauzy movability with a specified winner)

        A quadratic (or generalized) permutation is rauzy_movable type
        depending on the possible length of the last interval. It's
        dependent of the length equation.

        INPUT:

        - ``winner`` - 'top' (or 't' or 0) or 'bottom' (or 'b' or 1)

        OUTPUT:

        bool -- True if self has a Rauzy move

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.GeneralizedPermutation('a a','b b')
            sage: p.has_right_rauzy_move('top')
            False
            sage: p.has_right_rauzy_move('bottom')
            False

        ::

            sage: p = iet.GeneralizedPermutation('a a b','b c c')
            sage: p.has_right_rauzy_move('top')
            True
            sage: p.has_right_rauzy_move('bottom')
            True

        ::

            sage: p = iet.GeneralizedPermutation('a a','b b c c')
            sage: p.has_right_rauzy_move('top')
            True
            sage: p.has_right_rauzy_move('bottom')
            False

        ::

            sage: p = iet.GeneralizedPermutation('a a b b','c c')
            sage: p.has_right_rauzy_move('top')
            False
            sage: p.has_right_rauzy_move('bottom')
            True
        """
        winner = interval_conversion(winner)
        loser = self._labels[1-winner][-1]

        # the same letter at the right-end (False)
        if self._labels[0][-1] == self._labels[1][-1] :
            return False

        # the winner (or loser) letter is repeated on the other interval (True)
        if self._labels[0][-1] in self._labels[1]: return True
        if self._labels[1][-1] in self._labels[0]: return True

        # the loser letters is the only letter repeated in the loser
        # interval (False)
        for i,c in enumerate((self._labels[1-winner])):
            if c != loser and c in self._labels[1-winner][i+1:]:
                return True

        return False

    def right_rauzy_move(self, winner):
        r"""
        Perform a Rauzy move on the right (the standard one).

        INPUT:

        - ``winner`` - 'top' (or 't' or 0) or 'bottom' (or 'b' or 1)

        OUTPUT:

        boolean -- True if self has a Rauzy move

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.GeneralizedPermutation('a a b','b c c')
            sage: p.right_rauzy_move(0)
            a a b
            b c c
            sage: p.right_rauzy_move(1)
            a a
            b b c c

        ::

            sage: p = iet.GeneralizedPermutation('a b b','c c a')
            sage: p.right_rauzy_move(0)
            a a b b
            c c
            sage: p.right_rauzy_move(1)
            a b b
            c c a

        TESTS::

            sage: p = iet.GeneralizedPermutation('a a b','b c c')
            sage: q = p.top_bottom_inverse()
            sage: q = q.right_rauzy_move(0)
            sage: q = q.top_bottom_inverse()
            sage: q == p.right_rauzy_move(1)
            True
            sage: q = p.top_bottom_inverse()
            sage: q = q.right_rauzy_move(1)
            sage: q = q.top_bottom_inverse()
            sage: q == p.right_rauzy_move(0)
            True
            sage: p = p.left_right_inverse()
            sage: q = q.left_rauzy_move(0)
            sage: q = q.left_right_inverse()
            sage: q == p.right_rauzy_move(0)
            True
            sage: q = p.left_right_inverse()
            sage: q = q.left_rauzy_move(1)
            sage: q = q.left_right_inverse()
            sage: q == p.right_rauzy_move(1)
            True
        """
        result = copy(self)

        winner_letter = result._labels[winner][-1]
        loser_letter = result._labels[1-winner].pop(-1)

        if winner_letter in result._labels[winner][:-1]:
            loser_to = result._labels[winner].index(winner_letter)
            result._labels[winner].insert(loser_to, loser_letter)
        else:
            loser_to = result._labels[1-winner].index(winner_letter) + 1
            result._labels[1-winner].insert(loser_to, loser_letter)

        return result

    def left_rauzy_move(self, winner):
        r"""
        Perform a Rauzy move on the left.

        INPUT:

        - ``winner`` - 'top' or 'bottom'

        OUTPUT:

        permutation -- the Rauzy move of self

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.GeneralizedPermutation('a a b','b c c')
            sage: p.left_rauzy_move(0)
            a a b b
            c c
            sage: p.left_rauzy_move(1)
            a a b
            b c c

        ::

            sage: p = iet.GeneralizedPermutation('a b b','c c a')
            sage: p.left_rauzy_move(0)
            a b b
            c c a
            sage: p.left_rauzy_move(1)
            b b
            c c a a


        TESTS::

            sage: p = iet.GeneralizedPermutation('a a b','b c c')
            sage: q = p.top_bottom_inverse()
            sage: q = q.left_rauzy_move(0)
            sage: q = q.top_bottom_inverse()
            sage: q == p.left_rauzy_move(1)
            True
            sage: q = p.top_bottom_inverse()
            sage: q = q.left_rauzy_move(1)
            sage: q = q.top_bottom_inverse()
            sage: q == p.left_rauzy_move(0)
            True
            sage: q = p.left_right_inverse()
            sage: q = q.right_rauzy_move(0)
            sage: q = q.left_right_inverse()
            sage: q == p.left_rauzy_move(0)
            True
            sage: q = p.left_right_inverse()
            sage: q = q.right_rauzy_move(1)
            sage: q = q.left_right_inverse()
            sage: q == p.left_rauzy_move(1)
            True
        """
        result = copy(self)

        winner_letter = result._labels[winner][0]
        loser_letter = result._labels[1-winner].pop(0)

        if winner_letter in result._labels[winner][1:]:
            loser_to = result._labels[winner][1:].index(winner_letter)+2
            result._labels[winner].insert(loser_to, loser_letter)

        else:
            loser_to = result._labels[1-winner].index(winner_letter)
            result._labels[1-winner].insert(loser_to, loser_letter)

        return result

    def reduced(self):
        r"""
        Returns the associated reduced quadratic permutations.

        OUTPUT:

        permutation -- the underlying reduced permutation

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.GeneralizedPermutation('a a','b b c c')
            sage: q = p.reduced()
            sage: q
            a a
            b b c c
            sage: p.rauzy_move(0).reduced() == q.rauzy_move(0)
            True
        """
        from .reduced import ReducedPermutationLI

        return ReducedPermutationLI(self.list(),alphabet=self._alphabet, reduced=True)

    def rauzy_diagram(self, **kargs):
        r"""
        Returns the associated RauzyDiagram.

        OUTPUT:

        Rauzy diagram -- the Rauzy diagram of the permutation

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.GeneralizedPermutation('a b c b', 'c d d a')
            sage: d = p.rauzy_diagram()
            sage: p in d
            True

        For more information, try help(iet.RauzyDiagram)
        """
        return LabelledRauzyDiagram(self, **kargs)

    def lyapunov_exponents_H_plus(self, nb_vectors=None, nb_experiments=10,
                                  nb_iterations=65536, return_speed=False,
                                  verbose=False, output_file=None):
        r"""
        Compute the H^+ Lyapunov exponents of the stratum associated to this
        permutation.

        This method calls a C library. It might be  significantly faster if
        ``nb_vectors=1`` (or if it is not provided but genus is 1).

        INPUT:

        - ``nb_vectors`` -- the number of exponents to compute. The number of
          vectors must not exceed the dimension of the space!

         - ``nb_experiments`` -- the number of experiments to perform. It might
           be around 100 (default value) in order that the estimation of
           confidence interval is accurate enough.

         - ``nb_iterations`` -- the number of iteration of the Rauzy-Zorich
           algorithm to perform for each experiments. The default is 2^15=32768
           which is rather small but provide a good compromise between speed and
           quality of approximation.

        - ``verbose`` -- if ``True`` provide additional informations rather than
          returning only the Lyapunov exponents (i.e. ellapsed time, confidence
          intervals, ...)

        - ``output_file`` -- if provided (as a file object or a string) output
          the additional information in the given file rather than on the
          standard output.

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: Q = QuadraticStratum([1,1,-1,-1]).unique_component()
            sage: p = Q.permutation_representative(reduced=False)
            sage: p.lyapunov_exponents_H_plus(nb_iterations=2**20) # abs tol .05
            [0.6666]

            sage: Q_reg = QuadraticStratum([12]).regular_component()
            sage: p_reg = Q_reg.permutation_representative(reduced=False)
            sage: p_reg.lyapunov_exponents_H_plus(nb_iterations=2**20) # abs tol .05
            [0.662, 0.448, 0.230, 0.087]
            sage: sum(_)  # abs tol .05
            1.428

            sage: Q_irr = QuadraticStratum([12]).irregular_component()
            sage: p_irr = Q_irr.permutation_representative(reduced=False)
            sage: p_irr.lyapunov_exponents_H_plus(nb_iterations=2**20) # abs tol .05
            [0.747, 0.491, 0.245, 0.090]
            sage: sum(_) # abs tol .05
            1.572
        """
        if self._flips:
            raise NotImplementedError("Lyapunov exponents not implemented for permutations with flips")
        c = self.cover([[0]]*len(self), as_tuple=True)
        return c.lyapunov_exponents_H_plus(
                    nb_vectors=nb_vectors, nb_experiments=nb_experiments,
                    nb_iterations=nb_iterations, return_speed=return_speed,
                    verbose=verbose, output_file=output_file)

    def lyapunov_exponents_H_minus(self, nb_vectors=None, nb_experiments=10,
                                  nb_iterations=65536, return_speed=False,
                                  verbose=False, output_file=None):
        r"""
        Compute the H^+ Lyapunov exponents of the stratum associated to this
        permutation.

        This method calls a C library. It might be significantly faster if
        ``nb_vectors=1`` (or if it is not provided but genus is 1).

        INPUT:

        - ``nb_vectors`` -- the number of exponents to compute. The number of
          vectors must not exceed the dimension of the space!

         - ``nb_experiments`` -- the number of experiments to perform. It might
           be around 100 (default value) in order that the estimation of
           confidence interval is accurate enough.

         - ``nb_iterations`` -- the number of iteration of the Rauzy-Zorich
           algorithm to perform for each experiments. The default is `2^16=65536`
           which is rather small but provide a good compromise between speed and
           quality of approximation.

        - ``verbose`` -- if ``True`` provide additional informations rather than
          returning only the Lyapunov exponents (i.e. ellapsed time, confidence
          intervals, ...)

        - ``output_file`` -- if provided (as a file object or a string) output
          the additional information in the given file rather than on the
          standard output.

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: Q = QuadraticStratum([1,1,-1,-1]).unique_component()
            sage: p = Q.permutation_representative(reduced=False)
            sage: p.lyapunov_exponents_H_minus(nb_iterations=2**20) # abs tol .05
            [1.000, 0.333]

            sage: Q_reg = QuadraticStratum([12]).regular_component()
            sage: p_reg = Q_reg.permutation_representative(reduced=False)
            sage: p_reg.lyapunov_exponents_H_minus(nb_iterations=2**19) # abs tol .05
            [1.000, 0.309, 0.119]
            sage: sum(_)  # abs tol .05
            1.428

            sage: Q_irr = QuadraticStratum([12]).irregular_component()
            sage: p_irr = Q_irr.permutation_representative(reduced=False)
            sage: p_irr.lyapunov_exponents_H_minus(nb_iterations=2**19) # abs tol .05
            [1.000, 0.444, 0.128]
            sage: sum(_) # abs tol .05
            1.5725
        """
        if self._flips:
            raise NotImplementedError("Lyapunov exponents not implemented for permutations with flips")

        # we know that the double cover gives rise to two characters. We need to
        # find the one corresponding to H^-. We just pick the one which is not
        # constantly 1 and correspond to H^+.
        c = self.orientation_cover()
        c0,c1 = c._real_characters()[0]
        i0 = (-1 in c0)
        i1 = (-1 in c1)
        if i0 and i1:
            raise RuntimeError("not a generalized permutation")
        elif i0:
            character = c0
        elif i1:
            character = c1
        else:
            raise RuntimeError("trouble with permutation={}".format(self))

        return c.lyapunov_exponents_H_plus(
                    nb_vectors=nb_vectors, nb_experiments=nb_experiments,
                    nb_iterations=nb_iterations, return_speed=return_speed,
                    isotypic_decomposition=character,
                    verbose=verbose, output_file=output_file)


class FlippedLabelledPermutationIET(FlippedPermutationIET, LabelledPermutationIET):
    r"""
    Flipped labelled permutation from iet.

    EXAMPLES::

        sage: from surface_dynamics import *

    Reducibility testing (does not depends of flips)::

        sage: p = iet.Permutation('a b c', 'c b a',flips='a')
        sage: p.is_irreducible()
        True
        sage: q = iet.Permutation('a b c d', 'b a d c', flips='bc')
        sage: q.is_irreducible()
        False

    Rauzy movability and Rauzy move::

        sage: p = iet.Permutation('a b c', 'c b a',flips='a')
        sage: p
        -a  b  c
         c  b -a
        sage: p.rauzy_move(1)
        -c -a  b
        -c  b -a
        sage: p.rauzy_move(0)
        -a  b  c
         c -a  b

    Rauzy diagrams::

        sage: d = iet.RauzyDiagram('a b c d','d a b c',flips='a')
    """
    def reduced(self):
        r"""
        The associated reduced permutation.

        OUTPUT:

        permutation -- the associated reduced permutation

        EXAMPLE::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b c','c b a',flips='a')
            sage: q = iet.Permutation('a b c','c b a',flips='a',reduced=True)
            sage: p.reduced() == q
            True
        """
        from surface_dynamics.interval_exchanges.reduced import FlippedReducedPermutationIET

        return FlippedReducedPermutationIET(
            intervals=self.list(flips=False),
            flips=self.flips(),
            alphabet=self.alphabet(),
            reduced=True)

    def rauzy_diagram(self, **kargs):
        r"""
        Returns the Rauzy diagram associated to this permutation.

        For more information, try help(iet.RauzyDiagram)

        OUTPUT:

        RauzyDiagram -- the Rauzy diagram of self

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b c', 'c b a',flips='a')
            sage: p.rauzy_diagram()
            Rauzy diagram with 3 permutations
        """
        return FlippedLabelledRauzyDiagram(self, **kargs)

class FlippedLabelledPermutationLI(FlippedPermutationLI, LabelledPermutationLI):
    r"""
    Flipped labelled quadratic (or generalized) permutation.

    EXAMPLES::

            sage: from surface_dynamics import *

    Rauzy movability and Rauzy move::

        sage: p = iet.GeneralizedPermutation('a a b b c c', 'd d', flips='d')
        sage: p.has_rauzy_move(0)
        False
        sage: p.has_rauzy_move(1)
        True
        sage: p = iet.GeneralizedPermutation('a a b','b c c',flips='c')
        sage: p.has_rauzy_move(0)
        True
        sage: p.has_rauzy_move(1)
        True
    """
    def reduced(self):
        r"""
        The associated reduced permutation.

        OUTPUT:

        permutation -- the associated reduced permutation

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.GeneralizedPermutation('a a','b b c c',flips='a')
            sage: q = iet.GeneralizedPermutation('a a','b b c c',flips='a',reduced=True)
            sage: p.reduced() == q
            True
        """
        from surface_dynamics.interval_exchanges.reduced import FlippedReducedPermutationLI

        return FlippedReducedPermutationLI(
            intervals=self.list(flips=False),
            flips=self.flips(),
            alphabet=self.alphabet(),
            reduced=True)

    def right_rauzy_move(self, winner):
        r"""
        Perform a Rauzy move on the right (the standard one).

        INPUT:

        - ``winner`` - either 'top' or 'bottom' ('t' or 'b' for short)

        OUTPUT:

        permutation -- the Rauzy move of self

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.GeneralizedPermutation('a a b','b c c',flips='c')
            sage: p.right_rauzy_move(0)
             a  a  b
            -c  b -c
            sage: p.right_rauzy_move(1)
             a  a
            -b -c -b -c

        ::

            sage: p = iet.GeneralizedPermutation('a b b','c c a',flips='ab')
            sage: p.right_rauzy_move(0)
             a -b  a -b
             c  c
            sage: p.right_rauzy_move(1)
             b -a  b
             c  c -a
        """
        result = copy(self)

        winner_letter = result._labels[winner][-1]
        winner_flip = result._flips[winner][-1]

        loser_letter = result._labels[1-winner].pop(-1)
        loser_flip = result._flips[1-winner].pop(-1)

        if loser_letter in result._labels[winner]:
            loser_twin = result._labels[winner].index(loser_letter)
            result._flips[winner][loser_twin] = loser_flip*winner_flip
        else:
            loser_twin = result._labels[1-winner].index(loser_letter)
            result._flips[1-winner][loser_twin] = loser_flip*winner_flip

        if winner_letter in result._labels[winner][:-1]:
            loser_to = result._labels[winner].index(winner_letter)
            if winner_flip == -1: loser_to += 1
            result._labels[winner].insert(loser_to, loser_letter)
            result._flips[winner].insert(loser_to, loser_flip*winner_flip)
        else:
            loser_to = result._labels[1-winner].index(winner_letter)
            if loser_flip == 1: loser_to += 1
            result._labels[1-winner].insert(loser_to, loser_letter)
            result._flips[1-winner].insert(loser_to, loser_flip*winner_flip)

        return result

    def left_rauzy_move(self, winner):
        r"""
        Perform a Rauzy move on the left.

        INPUT:

        - ``winner`` - either 'top' or 'bottom' ('t' or 'b' for short)

        OUTPUT:

        -- a permutation

        EXAMPLES::

            sage: from surface_dynamics import *

        ::

            sage: p = iet.GeneralizedPermutation('a a b','b c c')
            sage: p.left_rauzy_move(0)
            a a b b
            c c
            sage: p.left_rauzy_move(1)
            a a b
            b c c

        ::

            sage: p = iet.GeneralizedPermutation('a b b','c c a')
            sage: p.left_rauzy_move(0)
            a b b
            c c a
            sage: p.left_rauzy_move(1)
            b b
            c c a a
        """
        result = copy(self)

        winner_letter = result._labels[winner][0]
        loser_letter = result._labels[1-winner].pop(0)

        if winner_letter in result._labels[winner][1:]:
            loser_to = result._labels[winner][1:].index(winner_letter)+2
            result._labels[winner].insert(loser_to, loser_letter)

        else:
            loser_to = result._labels[1-winner].index(winner_letter)
            result._labels[1-winner].insert(loser_to, loser_letter)

        return result

    def rauzy_diagram(self, **kargs):
        r"""
        Returns the associated Rauzy diagram.

        For more information, try help(RauzyDiagram)

        OUTPUT :

        -- a RauzyDiagram

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.GeneralizedPermutation('a b b a', 'c d c d')
            sage: d = p.rauzy_diagram()
        """
        return FlippedLabelledRauzyDiagram(self, **kargs)

class LabelledRauzyDiagram(RauzyDiagram):
    r"""
    Template for Rauzy diagrams of labelled permutations.

        ...DO NOT USE...
    """
    class Path(RauzyDiagram.Path):
        r"""
        Path in Labelled Rauzy diagram.
        """
        def matrix(self):
            r"""
            Returns the matrix associated to a path.

            The matrix associated to a Rauzy induction, is the linear
            application that allows to recover the lengths of self from the
            lengths of the induced.

            OUTPUT:

            matrix -- a square matrix of integers

            EXAMPLES::

                sage: from surface_dynamics import *

                sage: p = iet.Permutation('a1 a2','a2 a1')
                sage: d = p.rauzy_diagram()
                sage: g = d.path(p,'top')
                sage: g.matrix()
                [1 0]
                [1 1]
                sage: g = d.path(p,'bottom')
                sage: g.matrix()
                [1 1]
                [0 1]

            ::

                sage: p = iet.Permutation('a b c','c b a')
                sage: d = p.rauzy_diagram()
                sage: g = d.path(p)
                sage: g.matrix() == identity_matrix(3)
                True
                sage: g = d.path(p,'top')
                sage: g.matrix()
                [1 0 0]
                [0 1 0]
                [1 0 1]
                sage: g = d.path(p,'bottom')
                sage: g.matrix()
                [1 0 1]
                [0 1 0]
                [0 0 1]
            """
            return self.composition(self._parent.edge_to_matrix)

        def interval_substitution(self):
            r"""
            Returns the substitution of intervals obtained.

            OUTPUT:

            WordMorphism -- the word morphism corresponding to the interval

            EXAMPLES::

                sage: from surface_dynamics import *

                sage: p = iet.Permutation('a b','b a')
                sage: r = p.rauzy_diagram()
                sage: p0 = r.path(p,0)
                sage: s0 = p0.interval_substitution()
                sage: print(s0)
                a->a, b->ba
                sage: p1 = r.path(p,1)
                sage: s1 = p1.interval_substitution()
                sage: print(s1)
                a->ab, b->b
                sage: (p0 + p1).interval_substitution() == s1 * s0
                True
                sage: (p1 + p0).interval_substitution() == s0 * s1
                True
            """
            return self.right_composition(self._parent.edge_to_interval_substitution)

        def orbit_substitution(self):
            r"""
            Returns the substitution on the orbit of the left extremity.

            OUTPUT:

            WordMorhpism -- the word morphism corresponding to the orbit

            EXAMPLES::

                sage: from surface_dynamics import *

                sage: p = iet.Permutation('a b','b a')
                sage: d = p.rauzy_diagram()
                sage: g0 = d.path(p,'top')
                sage: s0 = g0.orbit_substitution()
                sage: print(s0)
                a->ab, b->b
                sage: g1 = d.path(p,'bottom')
                sage: s1 = g1.orbit_substitution()
                sage: print(s1)
                a->a, b->ab
                sage: (g0 + g1).orbit_substitution() == s0 * s1
                True
                sage: (g1 + g0).orbit_substitution() == s1 * s0
                True
            """
            return self.composition(self._parent.edge_to_orbit_substitution)

        substitution = orbit_substitution  # standard name
        dual_substitution = interval_substitution  # standard name

        def is_full(self):
            r"""
            Tests the fullness.

            A path is full if all intervals win at least one time.

            OUTPUT:

            boolean -- True if the path is full and False else

            EXAMPLE::

                sage: from surface_dynamics import *

                sage: p = iet.Permutation('a b c','c b a')
                sage: r = p.rauzy_diagram()
                sage: g0 = r.path(p,'t','b','t')
                sage: g1 = r.path(p,'b','t','b')
                sage: g0.is_full()
                False
                sage: g1.is_full()
                False
                sage: (g0 + g1).is_full()
                True
                sage: (g1 + g0).is_full()
                True
            """
            return set(self._parent.letters()) == set(self.winners())

        def self_similar_iet(self, name='a'):
            r"""
            Return the self-similar interval exchange transformation associated to this path

            INPUT:

            - ``name`` - an optional name for the generator of the number field

            EXAMPLES::

                sage: from surface_dynamics import *

            The golden rotation::

                sage: p = iet.Permutation('a b', 'b a')
                sage: R = p.rauzy_diagram()
                sage: g = R.path(p, 't', 'b')
                sage: T = g.self_similar_iet()
                sage: T.lengths().parent()
                Vector space of dimension 2 over Number Field ...
                sage: T.lengths().n()
                (1.00000000000000, 1.61803398874989)

            An example from Do-Schmidt::

                sage: code = [1,0,1,0,1,0,0,0,1,0,0,1,1,1,0,0,0,0,1,1,1,1,1,0,0,0,0,0,1,1,1,0]
                sage: p = iet.Permutation([0,1,2,3,4,5,6],[6,5,4,3,2,1,0])
                sage: R = p.rauzy_diagram()
                sage: g = R.path(p, *code)
                sage: T = g.self_similar_iet()
                sage: T.sah_arnoux_fathi_invariant()
                (0, 0, 0)
            """
            if not self.is_loop() or not self.is_full():
                raise ValueError("the path must be a full loop")

            from sage.rings.qqbar import AA
            from sage.rings.number_field.number_field import NumberField
            m = self.matrix()
            poly = m.charpoly()
            l = max(poly.roots(AA, False))
            K = NumberField(l.minpoly(), name=name, embedding=l)
            a = K.gen()
            lengths = (m - a).right_kernel().basis()[0]
            if any(x <= 0 for x in lengths):
                raise RuntimeError("wrong Perron-Frobenius eigenvector: {}".format(lengths))

            # NOTE: the above code makes "lengths" with parent being the right kernel (that is
            # a submodule of R^d)
            lengths = lengths.parent().ambient_vector_space()(lengths)

            from .iet import IntervalExchangeTransformation
            return IntervalExchangeTransformation(self.start(), lengths)

    def edge_to_interval_substitution(self, p=None, edge_type=None):
        r"""
        Returns the interval substitution associated to an edge

        OUTPUT:

        WordMorphism -- the WordMorphism corresponding to the edge

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b c','c b a')
            sage: r = p.rauzy_diagram()
            sage: print(r.edge_to_interval_substitution(None,None))
            a->a, b->b, c->c
            sage: print(r.edge_to_interval_substitution(p,0))
            a->a, b->b, c->ca
            sage: print(r.edge_to_interval_substitution(p,1))
            a->ac, b->b, c->c
        """
        if p is None and edge_type is None:
            return WordMorphism(dict((a,[a]) for a in self.letters()))

        function_name = self._edge_types[edge_type][0] + '_interval_substitution'
        if not hasattr(self._element_class,function_name):
            return WordMorphism(dict((a,[a]) for a in self.letters()))

        arguments = self._edge_types[edge_type][1]

        return getattr(p,function_name)(*arguments)

    def edge_to_orbit_substitution(self, p=None, edge_type=None):
        r"""
        Returns the interval substitution associated to an edge

        OUTPUT:

        WordMorphism -- the word morphism corresponding to the edge

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b c','c b a')
            sage: r = p.rauzy_diagram()
            sage: print(r.edge_to_orbit_substitution(None,None))
            a->a, b->b, c->c
            sage: print(r.edge_to_orbit_substitution(p,0))
            a->ac, b->b, c->c
            sage: print(r.edge_to_orbit_substitution(p,1))
            a->a, b->b, c->ac

        TESTS::

            sage: from surface_dynamics import *

            sage: pi0 = iet.Permutation('A1 A2 B', 'B A1 A2')
            sage: G = pi0.rauzy_diagram()
            sage: s1 = G.edge_to_orbit_substitution(pi0,0)
            sage: s1.domain().alphabet()
            {'A1', 'A2', 'B'}
            sage: s1.codomain().alphabet()
            {'A1', 'A2', 'B'}
        """
        if p is None and edge_type is None:
            return WordMorphism(dict((a,[a]) for a in self.letters()))

        function_name = self._edge_types[edge_type][0] + '_orbit_substitution'

        if not hasattr(self._element_class,function_name):
            return WordMorphism(dict((a,[a]) for a in self.letters()))

        arguments = self._edge_types[edge_type][1]
        return getattr(p,function_name)(*arguments)

    def full_loop_iterator(self, start=None, max_length=1):
        r"""
        Returns an iterator over all full path starting at start.

        INPUT:

        - ``start`` - the start point

        - ``max_length`` - a limit on the length of the paths

        OUTPUT:

        iterator -- iterator over full loops

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b','b a')
            sage: r = p.rauzy_diagram()
            sage: for g in r.full_loop_iterator(p,2):
            ....:     print("%s\n*****" % g.matrix())
            [1 1]
            [1 2]
            *****
            [2 1]
            [1 1]
            *****
        """
        g = self.path(start)

        ifull = filter(
            lambda x: x.is_loop() and x.is_full(),
            self._all_path_extension(g,max_length))

        return map(copy,ifull)

    def full_nloop_iterator(self, start=None, length=1):
        r"""
        Returns an iterator over all full loops of given length.

        INPUT:

        - ``start`` - the initial permutation

        - ``length`` - the length to consider

        OUTPUT:

        iterator -- an iterator over the full loops of given length

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b','b a')
            sage: d = p.rauzy_diagram()
            sage: for g in d.full_nloop_iterator(p,2):
            ....:     print("%s\n*****" % g.matrix())
            [1 1]
            [1 2]
            *****
            [2 1]
            [1 1]
            *****
        """
        g = self.path(start)

        ifull = filter(
            lambda x: x.is_loop() and x.is_full(),
            self._all_npath_extension(g,length))

        return map(copy, ifull)

    def _permutation_to_vertex(self, p):
        r"""
        Translation of a labelled permutation to a vertex

        INPUT:

        - ``p`` - a labelled Permutation

        TESTS::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b c','c b a')
            sage: r = p.rauzy_diagram()
            sage: p in r   #indirect doctest
            True
        """
        return (
        tuple(p._labels[0]),tuple(p._labels[1]),
        tuple(p._twin[0]),tuple(p._twin[1]))

    def _set_element(self,data):
        r"""
        Sets self._element with data

        TESTS::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b c','c b a')
            sage: r = p.rauzy_diagram()
            sage: r[p][0] == p.rauzy_move(0)   #indirect doctest
            True
            sage: r[p][1] == p.rauzy_move(1)   #indirect doctest
            True
        """
        self._element._labels = [list(data[0]), list(data[1])]
        self._element._twin = [list(data[2]), list(data[3])]

class FlippedLabelledRauzyDiagram(FlippedRauzyDiagram, LabelledRauzyDiagram):
    r"""
    Rauzy diagram of flipped labelled permutations
    """
    def _permutation_to_vertex(self, p):
        r"""
        Returns what must be stored from p.

        INPUT:

        - ``p`` - a Flipped labelled permutation

        TESTS::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b c','c b a',flips='a')
            sage: r = p.rauzy_diagram()
            sage: p in r   #indirect doctest
            True
        """
        return (tuple(p._labels[0]),tuple(p._labels[1]),
                tuple(p._twin[0]), tuple(p._twin[1]),
                tuple(p._flips[0]), tuple(p._flips[1]))

    def _set_element(self, data):
        r"""
        Returns what the vertex i as a permutation.

        TESTS::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b','b a',flips='a')
            sage: r = p.rauzy_diagram()
            sage: p in r   #indirect doctest
            True
        """
        self._element._labels = [list(data[0]), list(data[1])]
        self._element._twin = [list(data[2]), list(data[3])]
        self._element._flips = [list(data[4]), list(data[5])]




r"""
Constellations

A constellation is a tuple `(g_0, g_1, \dots, g_k)` of permutations
such that the product `g_0 g_1 ... g_k` is the identity. One often
assumes that the group generated by `g_0, g_1, \dots, g_k` acts
transitively ([LaZv04]_ definition 1). Geometrically, it corresponds
to a covering of the 2-sphere ramified over `k` points (the transitivity
condition corresponds to the connectivity of the covering).

.. NOTE::

    The permutations are on `[0, n-1]` and not on `[1, n]` for
    algorithmic and practical reasons.

EXAMPLES::

    sage: from surface_dynamics.misc.constellation import Constellation, Constellations
    sage: c = Constellation(['(0,1)', '(0,2)', None])
    sage: c
    Constellation of length 3 and degree 3
    g0 (0,1)(2)
    g1 (0,2)(1)
    g2 (0,2,1)
    sage: C = Constellations(3,4); C
    Connected constellations of length 3 and degree 4
    sage: C.cardinality()
    426

REFERENCES:

.. [LaZv04] S. Lando and A. Zvonkine, "Graphs on surfaces and their
            applications", Springer-Verlag, 2004.
"""
#*****************************************************************************
#       Copyright (C) 2010 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from __future__ import print_function, absolute_import, division
from six.moves import range, zip, filter

from sage.structure.parent import Parent
from sage.structure.element import Element
from sage.structure.unique_representation import UniqueRepresentation

from sage.rings.integer import Integer
from sage.combinat.partition import Partition
from sage.structure.sage_object import SageObject
from sage.groups.perm_gps.permgroup_element import PermutationGroupElement
from sage.graphs.graph import Graph

# we import everything from sage.misc.permutation
from surface_dynamics.misc.permutation import *

ZZ_0 = Integer(0)

# constructors

def Constellations(*data, **options):
    r"""
    Build a set of constellations.

    INPUT:

    - ``profile`` -- an optional profile

    - ``length`` -- an optional length

    - ``degree`` -- an optional degree

    - ``connected`` -- an optional boolean

    EXAMPLES::

        sage: from surface_dynamics.misc.constellation import Constellation, Constellations

    The set of all constellations::

        sage: Constellations()
        Connected constellations

    Constellations with constrained length and degree::

        sage: Constellations(4,2)
        Connected constellations of length 4 and degree 2
    """
    profile = options.get('profile', None)
    length = options.get('length', None)
    degree = options.get('degree', None)
    connected = options.get('connected', True)

    if data:
        if len(data) == 1:
            if isinstance(data[0],(tuple,list)):
                profile = data[0]
            else:
                length = Integer(data[0])
        elif len(data) == 2:
            length = Integer(data[0])
            degree = Integer(data[1])

    if profile:
        profile = tuple(map(Partition,profile))
        return Constellations_p(profile, bool(connected))
    if degree is None and length is None:
        return Constellations_all(bool(connected))
    elif degree is not None and length is not None:
        return Constellations_ld(Integer(length), Integer(degree), bool(connected))
    else:
        raise NotImplementedError("one cannot give just degree or just length")

def Constellation(g=None, mutable=False, connected=True, check=True):
    r"""
    Constellation

    INPUT:

    - ``g`` -- a list of permutations

    - ``mutable`` -- whether the result is mutable or not. Default is ``False``.

    - ``connected`` -- whether the result should be connected. Default is
      ``True``.

    - ``check`` -- whether or not to check. If it is ``True``, then the
      list g must contains no ``None``.

    EXAMPLES::

        sage: from surface_dynamics.misc.constellation import Constellation

    Simple initialization::

        sage: Constellation(['(0,1)','(0,3)(1,2)','(0,3,1,2)'])
        Constellation of length 3 and degree 4
        g0 (0,1)(2)(3)
        g1 (0,3)(1,2)
        g2 (0,3,1,2)

    One of the permutation can be omitted::

        sage: Constellation(['(0,1)', None, '(0,4)(1,2,3)'])
        Constellation of length 3 and degree 5
        g0 (0,1)(2)(3)(4)
        g1 (0,3,2,1,4)
        g2 (0,4)(1,2,3)

    One can define mutable constellations::

        sage: Constellation(([0,2,1], [2,1,0], [1,2,0]), mutable=True)
        Constellation of length 3 and degree 3
        g0 (0)(1,2)
        g1 (0,2)(1)
        g2 (0,1,2)
    """
    return Constellations(connected=connected)(g,check=check,mutable=mutable)

# classes

class Constellation_class(Element):
    r"""
    Constellation

    A constellation or a tuple of permutations `(g_0,g_1,...,g_k)`
    such that the product `g_0 g_1 ... g_k` is the identity.
    """
    def __init__(self, parent, g, check=True, connected=True, mutable=False):
        r"""
        TESTS::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation([[1,2,0],[0,2,1],[1,0,2],None])
            sage: c == loads(dumps(c))
            True
            sage: g0 = '(0,1)(2,4)'
            sage: g1 = '(0,3)(1,4)'
            sage: g2 = '(2,4,3)'
            sage: g3 = '(0,3)(1,2)'
            sage: c0 = Constellation([g0,g1,g2,g3])
            sage: c0 == Constellation([None,g1,g2,g3])
            True
            sage: c0 == Constellation([g0,None,g2,g3])
            True
            sage: c0 == Constellation([g0,g1,None,g3])
            True
            sage: c0 == Constellation([g0,g1,g2,None])
            True
        """
        Element.__init__(self, parent)
        self._connected = connected
        self._mutable = mutable
        self._g = g
        if check:
            self._check()

    def __hash__(self):
        r"""
        Return a hash for ``self``

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation(([0,2,1],[2,1,0],[1,2,0]), mutable=False)
            sage: c.__hash__() # random
            2148362403144019871
        """
        if self._mutable:
            raise ValueError("can not hash mutable constellation")
        return hash(tuple(map(tuple, self._g)))

    def set_immutable(self):
        r"""
        Does nothing, as ``self`` is already immutable

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation(([0,2,1],[2,1,0],[1,2,0]), mutable=False)
            sage: c.set_immutable()
            sage: c.is_mutable()
            False
        """
        self._mutable = False

    def is_mutable(self):
        r"""
        Return ``False`` as ``self`` is immutable

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation(([0,2,1],[2,1,0],[1,2,0]), mutable=False)
            sage: c.is_mutable()
            False
        """
        return self._mutable

    def switch(self, i, j0, j1):
        r"""
        Swith the vertices and edges at `i` and `i+1`

        The modification is local in the sense that it modifies `g_i`
        and `g_{i+1}` but does not modify the product `g_i g_{i+1}`.

        |  j0 -- g_i   --> k0 -- g_i0  --> l0
        |  j1 -- g_i+1 --> k1 -- g_i+1 --> l1

        becomes

        |  j0 -- g_i   --> k1 -- g_i   --> l0
        |  j1 -- g_i+1 --> k1 -- g_i+1 --> l1

        EXAMPLES::

           sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation(['(0,1)(2,3,4)','(1,4)',None], mutable=True); c
            Constellation of length 3 and degree 5
            g0 (0,1)(2,3,4)
            g1 (0)(1,4)(2)(3)
            g2 (0,1,3,2,4)
            sage: c.switch(1,2,3); c
            Constellation of length 3 and degree 5
            g0 (0,1)(2,3,4)
            g1 (0)(1,4)(2,3)
            g2 (0,1,3,4)(2)
        """
        if not self._mutable:
            raise ValueError("this constellation is immutable. Take a mutable copy first.")
        perm_switch(self._g[i], self._g[i+1], j0, j1)

    def _set_g_unsafe(self, i, j, k):
        r"""
        set gi[j] to k

        Advice: try a _check() after your modification

        TESTS::

            sage: from surface_dynamics import *
            sage: c = Constellation([[1,0],[0,1],[1,0]], mutable=True)
            sage: c
            Constellation of length 3 and degree 2
            g0 (0,1)
            g1 (0)(1)
            g2 (0,1)
            sage: c._set_g_unsafe(0,0,0)
            sage: c._set_g_unsafe(0,1,1)
            sage: c._set_g_unsafe(1,0,1)
            sage: c._set_g_unsafe(1,1,0)
            sage: c
            Constellation of length 3 and degree 2
            g0 (0)(1)
            g1 (0,1)
            g2 (0,1)
            sage: c._check()
        """
        if not self._mutable:
            raise ValueError("not mutable")
        self._g[i][j] = int(k)

    def euler_characteristic(self):
        r"""
        Return the Euler characteristic of the surface

        ALGORITHM:

        Hurwitz formula

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation(['(0,1)', '(0,2)', None])
            sage: c.euler_characteristic()
            2
        """
        return Integer(self.degree()*2 -
                       sum(sum(i-1 for i in self.profile(i))
                           for i in range(self.length())))

    def genus(self):
        r"""
        Returns the genus of the surface

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation
            sage: c = Constellation(['(0,1)', '(0,2)', None])
            sage: c.genus()
            0
        """
        return 1 - self.euler_characteristic()/2

    def _check(self):
        r"""
        Check that the constellation is valid and if not raise ValueError

        TESTS::

            sage: from surface_dynamics import *
            sage: c = Constellation([[1,0],[1,0]], mutable=True)
            sage: c._set_g_unsafe(0,0,0)
            sage: c._check()
            Traceback (most recent call last):
            ...
            ValueError: permutation 0 invalid
            sage: c._set_g_unsafe(0,1,1)
            sage: c._check()
            Traceback (most recent call last):
            ...
            ValueError: The product is not identity
            sage: c._set_g_unsafe(1,0,0)
            sage: c._set_g_unsafe(1,1,1)
            sage: c._check()
            Traceback (most recent call last):
            ...
            ValueError: not connected
        """
        for i in range(self.length()):
            if not perm_check(self._g[i]):
                raise ValueError("permutation {} invalid".format(i))

        h = perm_one(self.degree())
        for p in self._g:
            h = perm_compose(h, p)
        if not perm_is_one(h):
            raise ValueError("The product is not identity")

        if self._connected and not perms_are_transitive(self._g):
            raise ValueError("not connected")

    def __copy__(self):
        r"""
        TESTS::

            sage: from surface_dynamics import *
            sage: c = Constellation([[0,2,1],[1,0,2],[2,1,0],None])
            sage: c == copy(c)
            True
            sage: c is copy(c)
            False
            sage: c = Constellation([[0,2,1],[1,0,2],[2,1,0],None],mutable=True)
            sage: c == copy(c)
            True
            sage: c is copy(c)
            False
        """
        return self.parent()([gg[:] for gg in self._g],
                             check=False,
                             mutable=self._mutable)

    copy = __copy__

    def mutable_copy(self):
        r"""
        Return a mutable copy of ``self``

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation(([0,2,1],[2,1,0],[1,2,0]), mutable=False)
            sage: d = c.mutable_copy()
            sage: d.is_mutable()
            True
        """
        return self.parent()([gg[:] for gg in self._g],
                             check=False,
                             mutable=True)

    ## GENERAL PROPERTIES

    def is_connected(self):
        r"""
        Test of connectedness

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation(['(0,1)(2)', None, '(0,1)(2)'], connected=False)
            sage: c.is_connected()
            False
            sage: c = Constellation(['(0,1,2)', None], connected=False)
            sage: c.is_connected()
            True
        """
        if self._connected:
            return True
        else:
            return perms_are_transitive(self._g)

    def connected_components(self):
        """
        Returns the connected components ?

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation(['(0,1)(2)', None, '(0,1)(2)'], connected=False)
            sage: c.connected_components()
            [Constellation of length 3 and degree 2
             g0 (0,1)
             g1 (0)(1)
             g2 (0,1), Constellation of length 3 and degree 1
             g0 (0)
             g1 (0)
             g2 (0)]

            sage: c = Constellation(['(0,1,2)', None], connected=False)
            sage: c.connected_components()
            [Constellation of length 2 and degree 3
            g0 (0,1,2)
            g1 (0,2,1)]
        """
        if self._connected:
            return [self]
        G = Graph(loops=True, multiedges=True)
        for p in self._g:
            G.add_edges(enumerate(p))
        m = G.connected_components()
        if len(m) == 1:
            return [self]
        for mm in m:
            mm.sort()
        m.sort()
        g = [[] for _ in range(len(m))]
        m_inv = [None]*self.degree()
        for t in range(len(m)):
            for i in range(len(m[t])):
                m_inv[m[t][i]] = i
            for k in range(self.length()):
                tmp = [None]*len(m[t])
                for i in range(len(m[t])):
                    tmp[i] = m_inv[self._g[k][m[t][i]]]
                g[t].append(tmp)
        return [Constellation(g=g[i], check=False) for i in range(len(m))]

    def __eq__(self, other):
        r"""
        TESTS::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: Constellation(['(0,1,2)', None]) == Constellation(['(0,1,2)', None])
            True
            sage: Constellation(['(0,1)','(0,2)',None]) == Constellation(['(0,1)',None,'(0,2)'])
            False
        """
        return type(self) == type(other) and self._g == other._g

    def __ne__(self, other):
        r"""

        TESTS::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: Constellation(['(0,1,2)', None]) != Constellation(['(0,1,2)', None])
            False
            sage: Constellation(['(0,1)','(0,2)',None]) != Constellation(['(0,1)',None,'(0,2)'])
            True
        """
        return type(self) != type(other) or self._g != other._g

    def is_isomorphic(self, other, return_map=False):
        r"""
        Test of isomorphism

        Returns ``True`` if the constellations are isomorphic (i.e. common conjugacy)
        and returns the permutation that conjugate the two permutations if
        return_map is ``True`` in such a way that self.relabel(m) == other

        ALGORITHM:

            uses canonical labels obtained from the method :meth:`relabel`.

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation([[1,0,2],[2,1,0],[0,2,1],None])
            sage: d = Constellation([[2,1,0],[0,2,1],[1,0,2],None])
            sage: answer, mapping = c.is_isomorphic(d,return_map=True)
            sage: answer
            True
            sage: c.relabel(mapping) == d
            True
        """
        if return_map:
            if (self.degree() != other.degree() or
                self.length() != other.length()):
                return False, None
            sn, sn_map = self.relabel(return_map=True)
            on, on_map = other.relabel(return_map=True)
            if sn != on:
                return False, None
            return True, perm_compose(sn_map, perm_invert(on_map))

        return (self.degree() == other.degree() and
                self.length() == other.length() and
                self.relabel() == other.relabel())

    def __lt__(self, other):
        r"""
        TESTS::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c1 = Constellation([[1,2,0],None])
            sage: c2 = Constellation([[2,0,1],None])
            sage: c1 == c2
            False
            sage: c1 != c2
            True
            sage: c1 < c2 or c1 <= c2
            True
            sage: c1 > c2 or c1 >= c2
            False
            sage: c2 < c1 or c2 <= c1
            False
            sage: c2 > c1 and c2 >= c1
            True
        """
        if type(self) != type(other):
            raise TypeError

        if self.length() < other.length():
            return True
        elif self.length() > other.length():
            return False

        if self.degree() < other.degree():
            return True
        elif self.degree() > other.degree():
            return False

        for i in range(self.length()-1):
            for j in range(self.degree()-1):
                if self._g[i][j] < other._g[i][j]:
                    return True
                elif self._g[i][j] > other._g[i][j]:
                    return False

        # equality
        return False

    def __le__(self, other): return self == other or self < other
    def __ge__(self, other): return not self < other
    def __gt__(self, other): return not self <= other

    def _repr_(self):
        r"""
        String representation

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation([[1,0,2],[2,1,0],[0,2,1],None])
            sage: c._repr_()
            'Constellation of length 4 and degree 3\ng0 (0,1)(2)\ng1 (0,2)(1)\ng2 (0)(1,2)\ng3 (0,2)(1)'
        """
        s = "Constellation of length %d and degree %d" % (self.length(), self.degree())
        for i in range(self.length()):
            s += "\ng%d %s" % (i, perm_cycle_string(self._g[i], True))
        return s

    def degree(self):
        r"""
        Return the degree of the constellation

        The degree of a constellation is the number `n` that
        corresponds to the symmetric group `S(n)` in which the
        permutations of the constellation are defined.

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c=Constellation(['(0,1)',None])
            sage: c.degree()
            2
            sage: c=Constellation(['(0,1)','(0,3,2)(1,5)',None,'(4,3,2,1)'])
            sage: c.degree()
            6
        """
        return ZZ_0 if len(self._g) == 0 else Integer(len(self._g[0]))

    def length(self):
        r"""
        Number of permutations

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation(['(0,1)','(0,2)','(0,3)',None])
            sage: c.length()
            4
            sage: c = Constellation(['(0,1,3)',None,'(1,2)'])
            sage: c.length()
            3
        """
        return len(self._g)

    def profile(self, i=None):
        r"""
        Return the profile of ``self``

        The profile of a constellation is the tuple of partitions
        associated to the conjugacy classes of the permutations of the
        constellation.

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation(['(0,1,2)(3,4)','(0,3)',None])
            sage: c.profile()
            ([3, 2], [2, 1, 1, 1], [5])
        """
        if i is None:
            return tuple(self.profile(j) for j in range(self.length()))
        else:
            return Partition(sorted(map(len, perm_cycles(self._g[i], True)), reverse=True))

    passport = profile

    ## ACCESS TO INDIVIDUAL PERMUTATION

    def g(self, i=None):
        r"""
        Return the permutation i of the constellation as a list

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation(['(0,1,2)(3,4)','(0,3)',None])
            sage: c.g(0)
            [1, 2, 0, 4, 3]
            sage: c.g(1)
            [3, 1, 2, 0, 4]
            sage: c.g(2)
            [4, 0, 1, 2, 3]
        """
        from copy import copy
        if i is None:
            return list(map(copy, self._g))
        else:
            return copy(self._g[i])

    def g_cycle_tuples(self, i, singletons=False):
        r"""
        Return the tuples associated to the cycles of the permutations of ``self``

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation(['(0,1)(2,3,4)','(0,4)(1,3)',None])
            sage: c.g_cycle_tuples(0)
            [[0, 1], [2, 3, 4]]
            sage: c.g_cycle_tuples(1)
            [[0, 4], [1, 3]]
        """
        if i is None:
            return [perm_cycles(p, singletons) for p in self._g]
        else:
            return perm_cycles(self._g[i], singletons)

    def g_cycle_string(self, i, singletons=False):
        r"""
        Return the permutations of a constellation, as cycles

        If `i` is ``None``, return all of them

        Otherwise, return just the permutation `g_i` as a cycle

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation(['(0,1)(2,3,4)','(0,4)(1,3)',None])
            sage: c.g_cycle_string(None)
            ['(0,1)(2,3,4)', '(0,4)(1,3)', '(0,3)(1,2,4)']
            sage: c.g_cycle_string(2)
            '(0,3)(1,2,4)'
        """
        if i is None:
            return [perm_cycle_string(p, singletons) for p in self._g]
        else:
            return perm_cycle_string(self._g[i], singletons)

    def g_next(self, i, j):
        r"""
        Return the image of `j` for the permutation `g_i` in constant time

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation(['(1,2,0,5)','(2,3,4)(5,1)',None])
            sage: c.g_next(0,0)
            5
            sage: all(c.g(i)[j] == c.g_next(i,j) for i in range(3) for j in range(6))
            True
        """
        if i is None:
            return [p[j] for p in  self._g]
        else:
            return self._g[i][j]

    def g_prev(self, i, j):
        r"""
        Return the image of `j` for the permutation `g_i` in constant time

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation(['(0,1,2)(3,4)','(0,4,2)',None])
            sage: c.g_prev(0,0)
            2
            sage: all(c.g(i)[c.g_prev(i,j)] == j for i in range(3) for j in range(5))
            True
        """
        if i is None:
            return [self.g_prev(k, j) for k in range(self.length())]
        else:
            for k in range(i+1, self.length()):
                j = self._g[k][j]
            for k in range(0, i):
                j = self._g[k][j]
            return j

    def g_orbit(self, i, j):
        r"""
        Returns the orbit of `j` under `g_i`

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation(['(0,1)(2,3,4)','(1,4)',None])
            sage: c.g_orbit(0,3)
            [3, 4, 2]
            sage: c.g_orbit(1,0)
            [0]
        """
        if i is None:
            return [self.g_orbit(k, j) for k in range(self.length())]
        else:
            return perm_orbit(self._g[i], j)

    def relabel(self, perm=None, return_map=False):
        r"""
        Relabel ``self``.

        If ``perm`` is provided then relabel with respect to ``perm``. Otherwise
        use canonical labels. In that case, if ``return_map`` is provided, the
        return also the map used for canonical labels.

        Algorithm:

        the cycle for g(0) are adjacent and the cycle are joined with
        respect to the other permutations. The minimum is taken for
        all possible renumerotations.

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation(['(0,1)(2,3,4)','(1,4)',None]); c
            Constellation of length 3 and degree 5
            g0 (0,1)(2,3,4)
            g1 (0)(1,4)(2)(3)
            g2 (0,1,3,2,4)
            sage: c2 = c.relabel(); c2
            Constellation of length 3 and degree 5
            g0 (0,1)(2,3,4)
            g1 (0)(1,2)(3)(4)
            g2 (0,1,4,3,2)

        The map returned when the option ``return_map`` is set to True can be
        used to set the relabelling::

            sage: c3, perm = c.relabel(return_map=True)
            sage: c3 == c2 and c3 == c.relabel(perm=perm)
            True

            sage: d = c.relabel([4,3,1,0,2]); d
            Constellation of length 3 and degree 5
            g0 (0,2,1)(3,4)
            g1 (0)(1)(2,3)(4)
            g2 (0,1,2,4,3)
            sage: d.is_isomorphic(c)
            True

        We check that after a random relabelling the new constellation is
        isomorphic to the initial one::

            sage: c = Constellation(['(0,1)(2,3,4)','(1,4)',None])
            sage: p = SymmetricGroup(c.degree()).random_element()
            sage: cc = c.relabel([x-1 for x in p.domain()]).relabel()
            sage: cc.is_isomorphic(c)
            True
        """
        if perm is not None:
            perm.extend(range(len(perm), self.degree()))
            g = [[None]*self.degree() for _ in range(self.length())]
            for i in range(len(perm)):
                for k in range(self.length()):
                    g[k][perm[i]] = perm[self._g[k][i]]
            return Constellation(g=g, check=False, mutable=self.is_mutable())

        if return_map:
            try:
                return self._normal_form, self._normal_form_map
            except AttributeError:
                pass
        else:
            try:
                return self._normal_form
            except AttributeError:
                pass


        # compute canonical labels
        if not self.is_connected():
            raise ValueError("No canonical labels implemented for non connected constellation")

        c_win, m_win = perms_canonical_labels(self._g)
        c_win = self.parent()(c_win, mutable=False, check=False)

        if not self.is_mutable():
            self._normal_form = c_win
            self._normal_form_map = m_win

        c_win._normal_form = c_win
        c_win._normal_form_map = m_win

        if return_map:
            return c_win, m_win
        else:
            return c_win

    # BRAID GROUP ACTION

    def braid_group_action(self, i):
        r"""
        Acts on self as the braid group generator that exchanges
        position `i` and `i+1`

        INPUT:

        - ``i`` -- integer in `[0, n-1]` where `n` is the length of self

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: sigma = lambda c, i: c.braid_group_action(i)

            sage: c = Constellation(['(0,1)(2,3,4)','(1,4)',None]); c
            Constellation of length 3 and degree 5
            g0 (0,1)(2,3,4)
            g1 (0)(1,4)(2)(3)
            g2 (0,1,3,2,4)
            sage: sigma(c, 1)
            Constellation of length 3 and degree 5
            g0 (0,1)(2,3,4)
            g1 (0,1,3,2,4)
            g2 (0,3)(1)(2)(4)

        Check the commutation relation::

            sage: c = Constellation(['(0,1)(2,3,4)','(1,4)','(2,5)(0,4)',None])
            sage: d = Constellation(['(0,1,3,5)','(2,3,4)','(0,3,5)',None])
            sage: c13 = sigma(sigma(c, 0), 2)
            sage: c31 = sigma(sigma(c, 2), 0)
            sage: c13 == c31
            True
            sage: d13 = sigma(sigma(d, 0), 2)
            sage: d31 = sigma(sigma(d, 2), 0)
            sage: d13 == d31
            True

        Check the braid relation::

            sage: c121 = sigma(sigma(sigma(c, 1), 2), 1)
            sage: c212 = sigma(sigma(sigma(c, 2), 1), 2)
            sage: c121 == c212
            True
            sage: d121 = sigma(sigma(sigma(d, 1), 2), 1)
            sage: d212 = sigma(sigma(sigma(d, 2), 1), 2)
            sage: d121 == d212
            True
        """
        if i < 0 or i >= self.length():
            raise ValueError("i should be between 0 and %d"%(self.length()-1))
        j = i+1
        if j == self.length():   # wrap around the cylinder
            j = 0
        h = self.copy()
        si = self._g[i]
        sj = self._g[j]
        h._g[i] = sj
        h._g[j] = perm_compose(perm_compose(perm_invert(sj), si), sj)
        return h

    def braid_group_orbit(self):
        r"""
        Returns the graph of the action of the braid group.

        The action is considered up to isomorphism of constellation.

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: c = Constellation(['(0,1)(2,3,4)','(1,4)',None]); c
            Constellation of length 3 and degree 5
            g0 (0,1)(2,3,4)
            g1 (0)(1,4)(2)(3)
            g2 (0,1,3,2,4)
            sage: G = c.braid_group_orbit()
            sage: G.num_verts()
            4
            sage: G.num_edges()
            12
        """
        from sage.graphs.digraph import DiGraph

        G = DiGraph(multiedges=True, loops=True)
        waiting = [self.relabel()]

        while waiting:
            c = waiting.pop()
            G.add_vertex(c)
            for i in range(self.length()):
                cc = self.braid_group_action(i).relabel()
                if cc not in G:
                    waiting.append(cc)
                G.add_edge(c, cc, i)
        return G

class Constellations_all(UniqueRepresentation, Parent):
    Element = Constellation_class

    def __init__(self, connected=True):
        r"""
        TESTS::

            sage: from surface_dynamics.misc.constellation import Constellations
            sage: TestSuite(Constellations()).run()
        """
        from sage.categories.infinite_enumerated_sets import InfiniteEnumeratedSets
        Parent.__init__(self, category=InfiniteEnumeratedSets())
        self._connected = connected

    def __contains__(self,elt):
        r"""
        TESTS::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: C = Constellations(connected=True)
            sage: D = Constellations(connected=False)
            sage: C([[1,0],None]) in C
            True
            sage: C([[1,0],None]) in D
            True
            sage: D([[1,0],None]) in C
            True
            sage: D([[1,0],None]) in D
            True
            sage: D([[0,1],None]) in C
            False
            sage: D([[0,1],None]) in D
            True
        """
        return isinstance(elt, Constellation_class) and (not self._connected or elt.is_connected())

    def _repr_(self):
        r"""
        String representation.

        TESTS::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: Constellations()    # indirect doctest
            Connected constellations
        """
        if self._connected:
            return "Connected constellations"
        else:
            return "Constellations"

    def _element_constructor_(self, g, check=True, mutable=False):
        r"""
        Returns a constellation.

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellation

            sage: Constellation([[1,0],[1,0]])  # indirect doctest
            Constellation of length 2 and degree 2
            g0 (0,1)
            g1 (0,1)
        """
        if g is None or len(g) == 0:
            gg = []

        else:
            if g.count(None) == 0:
                i = None
            elif g.count(None) == 1:
                i = g.index(None)
                del g[i]
            else:
                raise ValueError("only one permutation must be None")

            gg = list(map(perm_init, g))
            equalize_perms(gg)

            if i is not None:
                h = range(len(gg[0]))
                for p in gg[i:]:
                    h = perm_compose(h, p)
                for p in gg[:i]:
                    h = perm_compose(h, p)
                gg.insert(i, perm_invert(h))

        return self.element_class(self, gg, check, self._connected, mutable)

    def an_element(self):
        r"""
        Return a constellation.

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: Constellations().an_element()
            Constellation of length 2 and degree 2
            g0 (0,1)
            g1 (0,1)
        """
        return Constellations(2,2).an_element()

    def __iter__(self):
        r"""
        TESTS::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: I = iter(Constellations())
            sage: for _ in range(100): print(next(I))
            Constellation of length 1 and degree 1
            g0 (0)
            Constellation of length 2 and degree 1
            g0 (0)
            g1 (0)
            Constellation of length 3 and degree 1
            g0 (0)
            g1 (0)
            g2 (0)
            Constellation of length 2 and degree 2
            g0 (0,1)
            g1 (0,1)
            ...
            Constellation of length 4 and degree 3
            g0 (0)(1,2)
            g1 (0)(1,2)
            g2 (0,2,1)
            g3 (0,1,2)
            Constellation of length 4 and degree 3
            g0 (0)(1,2)
            g1 (0)(1,2)
            g2 (0,2)(1)
            g3 (0,2)(1)
        """
        n = 2
        while True:
            for d in range(1,n):
                l = n-d
                for c in Constellations(l,d,connected=self._connected):
                    yield c
            n += 1

class Constellations_ld(UniqueRepresentation, Parent):
    r"""
    Constellations of given length and degree.

    EXAMPLES::

        sage: from surface_dynamics.misc.constellation import Constellations

        sage: C = Constellations(2,3); C
        Connected constellations of length 2 and degree 3
        sage: C([[1,2,0],[2,0,1]])
        Constellation of length 2 and degree 3
        g0 (0,1,2)
        g1 (0,2,1)
        sage: C.cardinality()
        2
        sage: Constellations(2,3,connected=False).cardinality()
        6
    """
    def __init__(self, length, degree, connected=True):
        """
        TESTS::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: TestSuite(Constellations(length=3,degree=4)).run()
        """
        from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
        Parent.__init__(self, facade=Constellations_all(), category=FiniteEnumeratedSets())
        self._length = Integer(length)
        self._degree = Integer(degree)
        if self._length <= 0:
            raise ValueError("length should be a positive integer")
        if self._degree <= 0:
            raise ValueError("degree should be a positive integer")
        self._connected = bool(connected)

    def __contains__(self, elt):
        r"""
        TESTS::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: C = Constellations(2,3,connected=True)
            sage: D = Constellations(2,3,connected=False)
            sage: C([[2,0,1],None]) in C
            True
            sage: D([[2,0,1],None]) in C
            True
            sage: D([[2,0,1],None]) in D
            True
            sage: D([[0,1,2],None]) in C
            False
            sage: D([[0,1,2],None]) in D
            True
        """
        if elt not in Constellations(connected=self._connected):
            return False
        return elt.length() == self._length and elt.degree() == self._degree

    def first(self):
        r"""
        Returns the first element in lexicographic order

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: const = Constellations(3,3);const
            Connected constellations of length 3 and degree 3
            sage: const.first()
            Constellation of length 3 and degree 3
            g0 (0,1,2)
            g1 (0,1,2)
            g2 (0,1,2)
        """
        if self._connected:
            g = [str(tuple(range(self._degree)))] * self._length
        else:
            g = [range(d)] * self._length
        return self(g, check=False)

    def last(self):
        r"""
        Returns the last element in lexicographic order

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: const=Constellations(3,3);const
            Connected constellations of length 3 and degree 3
            sage: const.last()
            Constellation of length 3 and degree 3
            g0 (0,2,1)
            g1 (0,2,1)
            g2 (0,2,1)
        """
        if self._connected:
            g = [str(tuple(range(self._degree-1, -1, -1)))] * self._length
        else:
            g = [range(d-1,-1,-1)] * self._length
        return Constellation(g, check=False, mutable=False)

    def _repr_(self):
        """
        TESTS::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: Constellations(3,3)._repr_()
            'Connected constellations of length 3 and degree 3'
            sage: Constellations(3,3,connected=False)._repr_()
            'Constellations of length 3 and degree 3'
        """
        s = "of length %d and degree %d" % (self._length, self._degree)
        if self._connected:
            return "Connected constellations " + s
        return "Constellations " + s

    def __iter__(self):
        """
        iterator over all constellations of given degree and length

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: const = Constellations(3,3); const
            Connected constellations of length 3 and degree 3
            sage: len([v for v in const])
            26

        One can check the first few terms of sequence :oeis:`220754`::

            sage: Constellations(4,1).cardinality()
            1
            sage: Constellations(4,2).cardinality()
            7
            sage: Constellations(4,3).cardinality()
            194
            sage: Constellations(4,4).cardinality()  # long time
            12858
        """
        from sage.arith.srange import srange
        from itertools import product, permutations

        if self._length == 1:
            if self._degree == 1:
                yield self([[0]])
            return

        for p in product(permutations(srange(self._degree)), repeat=self._length-1):
            if self._connected and not perms_are_transitive(p):
                continue
            yield self(list(map(list,p))+[None], check=False)

    def random_element(self, mutable=False):
        r"""
        a random element

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: const = Constellations(3,3)
            sage: const.random_element()
            Constellation of length 3 and degree 3
            ...
            ...
            ...
            sage: c = const.random_element()
            sage: c.degree() == 3 and c.length() == 3
            True
        """
        from sage.groups.perm_gps.permgroup_named import SymmetricGroup
        from sage.groups.perm_gps.permgroup import PermutationGroup

        l = self._length
        d = self._degree

        g = [SymmetricGroup(d).random_element() for _ in range(l-1)]
        G = PermutationGroup(g)
        while not G.degree() == d or (self._connected and not G.is_transitive()):
            g = [SymmetricGroup(d).random_element() for _ in range(l-1)]
            G = PermutationGroup(g)

        return self(list(map(permutation_to_perm, g))+[None], mutable=mutable)

    def _element_constructor_(self, *data, **options):
        r"""
        Build an element of self.

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: C = Constellations(2,3)
            sage: C([[1,2,0],[2,0,1]])
            Constellation of length 2 and degree 3
            g0 (0,1,2)
            g1 (0,2,1)
            sage: C([[2,1,0],[2,1,0]])
            Traceback (most recent call last):
            ...
            ValueError: not connected
        """
        c = Constellations(connected=self._connected)(*data, **options)
        if c.degree() != self._degree:
            raise ValueError("not able to build a constellation of degree %d from the given data"%self._degree)
        if c.length() != self._length:
            raise ValueError("not able to build a constellation of length %d from the given data"%self._length)
        return c

    def an_element(self):
        r"""
        Return a constellation in ``self``.

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: Constellations(2,3).an_element()
            Constellation of length 2 and degree 3
            g0 (0,2,1)
            g1 (0,1,2)
            sage: Constellations(7,3).an_element()
            Constellation of length 7 and degree 3
            g0 (0,2,1)
            g1 (0,1,2)
            g2 (0)(1)(2)
            g3 (0)(1)(2)
            g4 (0)(1)(2)
            g5 (0)(1)(2)
            g6 (0)(1)(2)
        """
        if self._degree == 0 and self._length == 0:
            return self()([])
        elif self._length == 1:
            if self._degree > 1:
                from sage.categories.sets_cat import EmptySetError
                raise EmptySetError
            return self()([[0]])

        d = self._degree
        if self._connected:
            g = [[d-1] + list(range(d-1)), list(range(1,d)) + [0]] + [list(range(d))]*(self._length-2)
        else:
            g = [list(range(d))]*self._length
        return self(g)

    def braid_group_action(self):
        r"""
        List of graphs that corresponds to the braid group action on ``self`` up
        to isomorphism.


        OUTPUT:

        - list of graphs

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: C = Constellations(3,3)
            sage: C.braid_group_action()
            [Looped multi-digraph on 3 vertices,
             Looped multi-digraph on 3 vertices,
             Looped multi-digraph on 1 vertex]
        """
        G = []
        I = iter(self)
        for c in self:
            c = c.relabel()
            if any(c in g for g in G):
                continue
            G.append(c.braid_group_orbit())
        return G

    def braid_group_orbits(self):
        r"""
        Orbits under the action of braid group.

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: C = Constellations(3,3)
            sage: O = C.braid_group_orbits()
            sage: len(O)
            3
            sage: [x.profile() for x in O[0]]
            [([1, 1, 1], [3], [3]), ([3], [1, 1, 1], [3]), ([3], [3], [1, 1, 1])]
            sage: [x.profile() for x in O[1]]
            [([2, 1], [2, 1], [3]), ([2, 1], [3], [2, 1]), ([3], [2, 1], [2, 1])]
            sage: [x.profile() for x in O[2]]
            [([3], [3], [3])]
        """
        return [g.vertices() for g in self.braid_group_action()]

class Constellations_p(UniqueRepresentation, Parent):
    r"""
    Constellations with fixed profile.

    EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellations

        sage: C = Constellations([[3,1],[3,1],[2,2]]); C
        Constellations with profile ([3, 1], [3, 1], [2, 2])
        sage: C.cardinality()
        24
        sage: C.first()
        Constellation of length 3 and degree 4
        g0 (0)(1,2,3)
        g1 (0,1,2)(3)
        g2 (0,1)(2,3)
        sage: C.last()
        Constellation of length 3 and degree 4
        g0 (0,3,2)(1)
        g1 (0,3,1)(2)
        g2 (0,1)(2,3)

    Note that the cardinality can also be computed using characters of the
    symmetric group (Frobenius formula)::

        sage: P = Partitions(4)
        sage: p1 = Partition([3,1])
        sage: p2 = Partition([3,1])
        sage: p3 = Partition([2,2])
        sage: i1 = P.cardinality() - P.rank(p1) - 1
        sage: i2 = P.cardinality() - P.rank(p2) - 1
        sage: i3 = P.cardinality() - P.rank(p3) - 1
        sage: s = 0
        sage: for c in SymmetricGroup(4).irreducible_characters():
        ....:     v = c.values()
        ....:     s += v[i1] * v[i2] * v[i3] / v[0]
        sage: c1 = p1.conjugacy_class_size()
        sage: c2 = p2.conjugacy_class_size()
        sage: c3 = p3.conjugacy_class_size()
        sage: c1 * c2 * c3 / factorial(4)**2 * s
        1

    The number obtained above is up to isomorphism. And we can check::

        sage: len(C.isomorphism_representatives())
        1
    """
    def __init__(self, profile, connected=True):
        r"""
        OPTIONS:

        - ``profile`` -- a list of integer partitions of the same integer

        - ``connected`` -- a boolean (default: True) that specify if we consider
          only connected constellations.

        TESTS::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: C = Constellations([(3,1),(3,1),(2,2)])
            sage: TestSuite(C).run()
        """
        l = Integer(len(profile))
        d = Integer(sum(profile[0]))
        for p in profile[1:]:
            if sum(p) != d:
                raise ValueError("all partition in the passport should have the same sum.")

        from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
        Parent.__init__(self, facade=Constellations_all(), category=FiniteEnumeratedSets())
        self._profile = profile
        self._degree = d
        self._length = l
        self._connected = connected

    def _repr_(self):
        r"""
        TESTS::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: Constellations(([3,2],[2,2,1]))
            Constellations with profile ([3, 2], [2, 2, 1])
        """
        return "Constellations with profile %s"%(self._profile,)

    def isomorphism_representatives(self):
        r"""
        Return a set of isomorphism representative of ``self``.

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: C = Constellations([[5], [4,1], [3,2]])
            sage: C.cardinality()
            240
            sage: ir = sorted(C.isomorphism_representatives())
            sage: len(ir)
            2
            sage: ir[0]
            Constellation of length 3 and degree 5
            g0 (0,1,2,3,4)
            g1 (0)(1,2,3,4)
            g2 (0,4,2)(1,3)
            sage: ir[1]
            Constellation of length 3 and degree 5
            g0 (0,1,2,3,4)
            g1 (0)(1,4,2,3)
            g2 (0,4)(1,2,3)
        """
        result = set()
        for c in self:
            cc = c.relabel()
            if cc not in result:
                result.add(cc)
        return result

    def _element_constructor_(self, *data, **options):
        r"""
        Build an element of self.

        TESTS::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: C = Constellations([(3,1),(3,1),(2,2)])
            sage: c = C(['(1,2,3)','(0,1,2)','(0,1)(2,3)']); c
            Constellation of length 3 and degree 4
            g0 (0)(1,2,3)
            g1 (0,1,2)(3)
            g2 (0,1)(2,3)
            sage: C(['(1,2,3)','(0,1,3)',None])
            Traceback (most recent call last):
            ...
            ValueError: not able to build a constellation with profile ([3, 1], [3, 1], [2, 2]) from the given data
        """
        c = Constellations(connected=self._connected)(*data, **options)
        if options.get('check', True):
            if c.profile() != self._profile:
                raise ValueError("not able to build a constellation with profile %s from the given data"%(self._profile,))
        return c

    def __iter__(self):
        r"""
        Iterator of the elements in ``self``.

        .. NOTE::

            The algorithm is highly non optimized and is *very* slow for large
            entries! In order to make a better one, we need an iterator over
            conjugacy classes of the symmetric group.

        TESTS::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: C = Constellations([(3,1),(3,1),(2,2)])
            sage: for c in C: print(c)
            Constellation of length 3 and degree 4
            g0 (0)(1,2,3)
            g1 (0,1,2)(3)
            g2 (0,1)(2,3)
            Constellation of length 3 and degree 4
            g0 (0)(1,2,3)
            g1 (0,2,3)(1)
            g2 (0,2)(1,3)
            ...
            Constellation of length 3 and degree 4
            g0 (0,3,2)(1)
            g1 (0,1,2)(3)
            g2 (0,3)(1,2)
            Constellation of length 3 and degree 4
            g0 (0,3,2)(1)
            g1 (0,3,1)(2)
            g2 (0,1)(2,3)
        """
        from sage.arith.srange import srange
        from itertools import product, permutations

        if self._length == 1:
            if self._degree == 1:
                yield self([[0]])
            return

        for p in product(permutations(srange(self._degree)), repeat=self._length-1):
            if self._connected and not perms_are_transitive(p):
                continue
            c = Constellations(connected=self._connected)(list(map(list,p))+[None], check=False)
            if c.profile() == self._profile:
                yield c

    def braid_group_action(self):
        r"""
        List of graphs that corresponds to the braid group action on ``self`` up
        to isomorphism.

        OUTPUT:

        - list of graphs

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: C = Constellations(3,3)
            sage: C.braid_group_action()
            [Looped multi-digraph on 3 vertices,
             Looped multi-digraph on 3 vertices,
             Looped multi-digraph on 1 vertex]
        """
        G = []
        I = iter(self)
        for c in self:
            c = c.relabel()
            if any(c in g for g in G):
                continue
            G.append(c.braid_group_orbit())
        return G

    def braid_group_orbits(self):
        r"""
        Orbits under the action of braid group.

        EXAMPLES::

            sage: from surface_dynamics.misc.constellation import Constellations

            sage: C = Constellations(3,3)
            sage: O = C.braid_group_orbits()
            sage: len(O)
            3
            sage: [x.profile() for x in O[0]]
            [([1, 1, 1], [3], [3]), ([3], [1, 1, 1], [3]), ([3], [3], [1, 1, 1])]
            sage: [x.profile() for x in O[1]]
            [([2, 1], [2, 1], [3]), ([2, 1], [3], [2, 1]), ([3], [2, 1], [2, 1])]
            sage: [x.profile() for x in O[2]]
            [([3], [3], [3])]
        """
        return [g.vertices() for g in self.braid_group_action()]




r"""
Topological recursion

This module implements topological recursion. In particular it makes
available the computation of various integral of cohomology classes
on M_{g,n} (psi, lambda and kappa classes).

REFS:

- G. Borot
  "Topological recursion and geometry"
  arXiv:1705.09986

- J. E. Andersen, G. Borot, L. O. Chekhov, N. Orantin
  "The ABCD of topological recursion"
  arXiv:1703.03307

- R. Pandharipande
  "Cohomological field theory calculations"
  arXiv:1712.02528

TODO:

- other recursions:
  - ELSV, Eynard Bouchard-Marino conjectures and simple Hurwitz numbers (lambda classes)
  - bundle of quadratic differentials and co
- generating series
- simplification of the recursion via the generalized string and dilaton
"""
#*****************************************************************************
#       Copyright (C) 2020 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************
from __future__ import print_function
from six.moves import range

from collections import defaultdict
from itertools import combinations

from sage.misc.misc_c import prod
from sage.misc.cachefunc import cached_method

from sage.rings.all import ZZ, QQ
from sage.arith.misc import factorial, binomial
from sage.combinat.integer_lists.invlex import IntegerListsLex
from sage.combinat.composition import Compositions
from sage.combinat.partition import Partitions

ZZ_0 = ZZ.zero()
ZZ_2 = ZZ(2)

def compositions(g, n):
    r"""
    Return compositions of sum at most 3g-3+n and length n
    """
    for k in range(0, 3*g-3+n+1):
        for c in Compositions(k+n, length=n, min_part=1):
            yield [i-1 for i in c]

class TopologicalRecursion:
    r"""
    A generic topological recursion.

    For concrete examples see:

    - :class:`~surface_dynamics.topological_recursion.kontsevich.KontsevichTR`
    - :class:`~surface_dynamics.topological_recursion.masur_veech.MasurVeechTR`
    - :class:`~surface_dynamics.topological_recursion.symbolic.SymbolicTR`
    """
    def __init__(self, cache_all=False, base_ring=None):
        self._cache = {}
        self._cache_all = cache_all
        self._base_ring = QQ if base_ring is None else base_ring
        self._one = None

    # The functions A, B, C, D of the topological recursion
    # They have to be implemented in subclasses
    def A(self, i, j, k): raise NotImplementedError

    def B(self, g, n, i, j): raise NotImplementedError
    # iterator (k, B(i, j, k))

    def C(self, g, n, i): raise NotImplementedError
    # iterator (j, k, C(i, j, k))

    def D(self, i): raise NotImplementedError

    def F(self, g, n, I):
        r"""
        Return the value of the partition function F_{g,n}(I) where I = (i_0, i_1, ..., i_{n-1}) is
        a tuple of n non-negative integers whose sum is <= 3g-3+n.

        We assume here that the topological recursion given by A, B, C and D produces a partition
        function symmetric in the variables i_0, i_1, ..., i_{n-1}.
        """
        # we restrict to TR with sum(I) <= 3g-3+n
        assert len(I) == n
        if len(I) != n or sum(I) > 3*g-3+n:
            raise ValueError("length must be n = %d and sum at most 3g-3+n = %d, got I = %s" % (n, 3*g-3+n, I))

        I = tuple(sorted(I))

        if g == 0 and n == 3:
            # M_{0,3} initial condition
            if self._one is None:
                self._one = self._base_ring(self.A(I[0], I[1], I[2]))
            return self._one
        elif g == 1 and n == 1:
            # M_{1,1} initial condition
            return self.D(I[0])

        return self._virasoro(g, n, I)

    def _virasoro(self, g, n, I):
        # TODO: we should use exponential notations for I
        if (g, n, I) in self._cache:
            return self._cache[(g, n, I)]

        Idict = defaultdict(int)
        for i in I[1:]:
            Idict[i] += 1
        Ituple = sorted(Idict.items())

        # set to True for debugging information
        verbose = False

        if verbose:
            print("Frec({}, {}, {}".format(g, n, I))

        B = self.B
        C = self.C
        F = self.F

        if verbose:
            print("compute S1")
        J = I[1:]
        S1 = ZZ_0
        for im,_ in Ituple:
            fac = Idict[im]
            J = []
            for j,mult in Ituple:
                if j == im:
                    J.extend([j] * (mult - 1))
                else:
                    J.extend([j] * mult)
            J = tuple(J)

            s = sum(J) # = i + j
            for (a, value) in B(g, n-s-1, I[0], im):
                if verbose:
                    print("[S1] B({}, {}, {}) F({}, {}, {})".format(I[0], im, a, g, n-1, (a,) + J))
                S1 += fac * value * F(g, n-1, (a,) + J)
        if verbose:
            print("S1 = {}".format(S1))

        if verbose:
            print("S2")
        S2 = ZZ_0
        if g >= 1:
            # TODO: here the bound is actually for the sum since the new tuple
            # is (a, b) + I[1:].
            I1 = I[1:]
            abbound = 3 * (g-1) - 3 + (n+1) - sum(I1)
            for (a, b, value) in C(I[0], abbound, abbound, abbound):
                if verbose:
                    print("[S2] C({}, {}, {}) F({}, {}, {})".format(I[0], a, b, g-1, n+1, (a,b) + I1))
                S2 += value * F(g-1, n+1, (a, b) + I1)
        if verbose:
            print("S2 = {}".format(S2))

        if verbose:
            print("S3")
        S3 = ZZ_0
        for n1 in range(n):
            n2 = n - n1 - 1

            g1min = int(n1 < 2)
            g2min = int(n2 < 2)

            for M1 in IntegerListsLex(min_sum=n1, max_sum=n1, length=len(Ituple), ceiling=[v for k,v in Ituple]):
                I1 = []
                I2 = []
                fac = 1
                for m,(k,v) in zip(M1, Ituple):
                    I1.extend([k]*m)
                    I2.extend([k]*(v-m))
                    fac *= binomial(v, m)
                I1 = tuple(I1)
                I2 = tuple(I2)

                s1 = sum(I1)
                s2 = sum(I2)

                # the above choices of combinations already forces the genus
                # ie, 3g1-3+n1 >= s1 and 3g2-3+n2 >= s2
                # g1 >= (s1 + 3 - n1)/3
                gg1min = max(g1min, (s1-n1+4)//3)
                gg2min = max(g2min, (s2-n2+4)//3)
                gg1max = g - gg2min
                for g1 in range(gg1min, gg1max+1):
                    g2 = g - g1
                    a1bound = 3*g1 - 3 + (n1+1) - s1
                    a2bound = 3*g2 - 3 + (n2+1) - s2
                    for (a1, a2, value) in C(I[0], a1bound, a2bound, a1bound + a2bound):
                        f1 = F(g1, n1 + 1, (a1,) + I1)
                        f2 = F(g2, n2 + 1, (a2,) + I2)
                        if verbose:
                            print("[S3] C({}, {}, {}) F({}, {}, {}) F({}, {}, {})".format(
                            I[0], a1, a2, g1, n1+1, (a1,) + I1, g2, n2+1, (a2,) + I2))
                        S3 += fac * value * f1 * f2

        if verbose:
            print("S3 = {}".format(S3))

        value = S1 + (S2 + S3) / ZZ_2
        if self._cache_all or I[0] >= 2:
            # We do not cache string/dilaton if _cache_all = False
            self._cache[(g, n, I)] = value
        return value

    def polynomial(self, g, n, ring=None):
        if ring is None:
            from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
            if n == 1:
                # Otherwise, the b is not numbered!
                ring = PolynomialRing(self._base_ring, 'b0', n)
            else:
                ring = PolynomialRing(self._base_ring, 'b', n)
            b = ring.gens()
        p = ring.zero()
        p += sum(self.F(g,n,I) * prod(b[i]**j for i,j in enumerate(I)) for I in compositions(g,n))
        return p

    def write(self, g, n, s=None):
        r"""
        Print normalized values of the polynomials F_{g,n}
        """
        g = ZZ(g)
        n = ZZ(n)
        if n < 0:
            raise ValueError
        if s is None:
            s = range(3*g-3+n+1)
        elif isinstance(s, numbers.Integral):
            s = [s]

        for t in s:
            for p in Partitions(t+n, length=n):
                p = [i-1 for i in p]
                c = prod(factorial(2*i+1) for i in p)
                f = self.F(g, n, p)
                if f:
                    print("{} {}".format([i for i in p], self.F(g, n, p) / c))

